import freeOrionAIInterface as fo

import AIDependencies
from buildings import BuildingType
from colonization.colony_score import debug_rating
from common.fo_typing import SystemId
from EnumsAI import FocusType
from freeorion_tools import (
    get_named_int,
    get_named_real,
    get_species_research,
    tech_soon_available,
)
from freeorion_tools.bonus_calculation import Bonus
from freeorion_tools.caching import cache_for_current_turn
from PlanetUtilsAI import get_empire_populated_planets
from turn_state import get_empire_planets_by_species, have_computronium


def calculate_research(planet: fo.planet, species: fo.species, max_population: float, stability: float) -> float:
    """
    Calculate how much PP the planet's population could generate with industry focus.
    This only considers values that actually rely on industry focus, those that do not are handled by
    calculate_planet_colonization_rating._rate_focus_independent.
    """
    if stability <= 0.0:
        return 0.0

    bonus_modified = _get_research_bonus_modified(planet, stability)
    flat_modified = _get_research_flat_modified(stability)
    skill_multiplier = get_species_research(species.name)
    bonus_by_policy = _get_research_bonus_modified_by_policy(stability)
    flat_by_policy = _get_research_flat_modified_by_policy(planet, stability)
    policy_multiplier = _get_policy_multiplier(stability)
    bonus_unmodified = _get_research_bonus_unmodified(planet, stability)

    result = (max_population * (AIDependencies.RESEARCH_PER_POP + bonus_modified) + flat_modified) * skill_multiplier
    result = (result + max_population * bonus_by_policy + flat_by_policy) * policy_multiplier
    result += max_population * bonus_unmodified
    debug_rating(
        f"calculate_research pop={max_population:.2f}, st={stability:.2f}, b1={bonus_modified:.2f}, "
        f"m1={skill_multiplier:.2f}, b2={bonus_by_policy:.2f}, f2={flat_by_policy:.2f}, "
        f"m2={policy_multiplier:.2f}, b3={bonus_unmodified:.2f} -> {result:.2f}"
    )
    return result


@cache_for_current_turn
def _get_modified_research_bonuses(planet: fo.planet) -> list[Bonus]:
    """
    Get list of per-population bonuses which are added before multiplication with the species skill value.
    """
    specials = planet.specials
    return [
        Bonus(
            AIDependencies.ANCIENT_RUINS_SPECIAL in specials or AIDependencies.ANCIENT_RUINS_SPECIAL2 in specials,
            get_named_real("ANCIENT_RUINS_MIN_STABILITY"),
            get_named_real("ANCIENT_RUINS_TARGET_RESEARCH_PERPOP"),
        ),
        Bonus(
            tech_soon_available("GRO_ENERGY_META", 3),
            get_named_real("GRO_ENERGY_META_MIN_STABILITY"),
            get_named_real("GRO_ENERGY_META_TARGET_RESEARCH_PERPOP"),
        ),
        Bonus(
            bool(BuildingType.ENCLAVE_VOID.built_or_queued_at()),
            get_named_real("BLD_ENCLAVE_VOID_MIN_STABILITY"),
            get_named_real("BLD_ENCLAVE_VOID_TARGET_RESEARCH_PERPOP"),
        ),
    ]


def _get_research_flat_modified(stability: float) -> float:
    """
    Get list of flat bonuses which are added before multiplication with the species skill value.
    """
    diversity = Bonus(
        fo.getEmpire().policyAdopted("PLC_DIVERSITY"),
        get_named_real("PLC_DIVERSITY_MIN_STABILITY"),
        (len(get_empire_planets_by_species()) - get_named_int("PLC_DIVERSITY_THRESHOLD"))
        * get_named_real("PLC_DIVERSITY_SCALING"),
    )
    return diversity.get_bonus(stability)


def _get_research_bonus_modified(planet: fo.planet, stability: float) -> float:
    """
    Calculate bonus research per population which would be added before multiplication with the species skill value.
    """
    return sum(bonus.get_bonus(stability) for bonus in _get_modified_research_bonuses(planet))


@cache_for_current_turn
def _get_modified_by_policy_research_bonuses(stability: float) -> list[Bonus]:
    return [
        Bonus(
            tech_soon_available("LRN_QUANT_NET", 3),
            get_named_real("LRN_QUANT_NET_MIN_STABILITY"),
            get_named_real("LRN_QUANT_NET_TARGET_RESEARCH_PERPOP"),
        ),
        # TBD check connection, _empire_resources should collect a list of systems or planets
        Bonus(
            have_computronium(),
            get_named_real("COMPUTRONIUM_MIN_STABILITY"),
            get_named_real("COMPUTRONIUM_TARGET_RESEARCH_PERPOP"),
        ),
        Bonus(
            fo.getEmpire().policyAdopted("PLC_LIBERTY"),
            get_named_real("PLC_LIBERTY_MIN_STABILITY"),
            (min(stability, get_named_real("PLC_LIBERTY_MAX_STABILITY")) - get_named_real("PLC_LIBERTY_MIN_STABILITY"))
            * get_named_real("PLC_LIBERTY_RESEARCH_BONUS_SCALING"),
        ),
    ]


def _get_research_bonus_modified_by_policy(stability: float) -> float:
    """
    Calculate bonus research per population which we would get independent of the species research skill,
    but still affected by industrialism.
    """
    return sum(bonus.get_bonus(stability) for bonus in _get_modified_by_policy_research_bonuses(stability))


@cache_for_current_turn
def _get_stellar_tomography_bonus(system_id: SystemId) -> float:
    universe = fo.getUniverse()
    system = universe.getSystem(system_id)
    if system.starType == fo.starType.blackHole:
        factor = get_named_real("LRN_STELLAR_TOMO_BLACK_TARGET_RESEARCH_PERPLANET")
    elif system.starType == fo.starType.neutron:
        factor = get_named_real("LRN_STELLAR_TOMO_NEUTRON_TARGET_RESEARCH_PERPLANET")
    else:
        factor = get_named_real("LRN_STELLAR_TOMO_NORMAL_STAR_TARGET_RESEARCH_PERPLANET")

    def is_our_researcher(pid):
        planet = universe.getPlanet(pid)
        return planet.focus == FocusType.FOCUS_RESEARCH and planet.owner == fo.empireID()

    num_researcher = sum(1 for pid in system.planetIDs if is_our_researcher(pid))
    # if we add another planet with research focus, all will profit. (n+1)^2 - n^2 = 2n + 1
    return (num_researcher * 2 + 1) * factor


def _get_research_flat_modified_by_policy(planet: fo.planet, stability: float) -> float:
    bonuses = [
        Bonus(
            tech_soon_available("LRN_STELLAR_TOMOGRAPHY", 3),
            0.0,
            _get_stellar_tomography_bonus(planet.systemID),
        ),
        # TODO Species InterDesign Academy
    ]
    return sum(bonus.get_bonus(stability) for bonus in bonuses)


def _get_policy_multiplier(stability) -> float:
    # if fo.getEmpire().policyAdopted("TECHNOCRACY") and stability >= get_named_real("PLC_TECHNOCRACY_MIN_STABILITY"):
    #     return 1.0 + get_named_real("PLC_TECHNOCRACY_TARGET_RESEARCH_PERCENT")
    if fo.getEmpire().policyAdopted("TECHNOCRACY") and stability >= get_named_real("PLC_TECHNOCRACY_MIN_STABILITY"):
        return 1.0 + get_named_real("PLC_TECHNOCRACY_TARGET_RESEARCH_PERCENT")
    return 1.0


@cache_for_current_turn
def _get_distributed_thought_bonus(system_id: SystemId) -> float:
    universe = fo.getUniverse()
    max_distance = max(
        (
            universe.linearDistance(system_id, p.systemID)
            for p in get_empire_populated_planets()
            if p.focus == FocusType.FOCUS_RESEARCH
        ),
        default=0.0,
    )
    # TBD? This could also affect other planets...
    return get_named_real("DISTRIB_THOUGH_RESEARCH_SCALING") * max_distance**0.5


def _get_research_bonus_unmodified(planet: fo.planet, stability: float):
    """
    Calculate bonus research per population which we would get independent of the species research skill.
    """
    bonuses = [
        Bonus(
            fo.getEmpire().policyAdopted("PLC_ALGORITHMIC_RESEARCH"),
            get_named_real("LRN_ALGO_RESEARCH_MIN_STABILITY"),
            get_named_real("LRN_ALGO_RESEARCH_TARGET_RESEARCH_PERCONSTRUCTION"),
        ),
        Bonus(
            tech_soon_available("LRN_DISTRIB_THOUGHT", 3),
            get_named_real("LRN_DISTRIB_THOUGHT_MIN_STABILITY"),
            _get_distributed_thought_bonus(planet.systemID),
        ),
        # TODO: distributed thought and collective thought
    ]
    return sum(bonus.get_bonus(stability) for bonus in bonuses)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_GRAVITONICS",
    description="LRN_GRAVITONICS_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=128 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_FORCE_FIELD"],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_STARLANE_BORE"),
        Item(type=UnlockBuilding, name="BLD_SCRYING_SPHERE"),
    ],
    graphic="icons/tech/gravitonics.png",
)



from focs._effects import (
    EffectsGroup,
    Focus,
    Happiness,
    NamedReal,
    NamedRealLookup,
    OwnedBy,
    Planet,
    SetMaxDefense,
    SetMaxFuel,
    SetMaxShield,
    SetTargetIndustry,
    SetTargetResearch,
    Ship,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import INDUSTRY_PER_POP, RESEARCH_PER_POP, TECH_COST_MULTIPLIER
from macros.misc import PLANET_DEFENSE_FACTOR, PLANET_SHIELD_FACTOR
from macros.priorities import (
    TARGET_AFTER_SCALING_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
)

Tech(
    name="GRO_ENERGY_META",
    description="GRO_ENERGY_META_DESC",
    short_description="VARIOUS_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=330 * TECH_COST_MULTIPLIER,
    researchturns=15,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=[
        "GRO_NANO_CYBERNET",
        "LRN_EVERYTHING",
        "GRO_TRANSORG_SENT",
    ],
    effectsgroups=[
        EffectsGroup(
            scope=Ship & OwnedBy(empire=Source.Owner),
            effects=SetMaxFuel(value=Value + 1),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Happiness(low=0),
            stackinggroup="DEFENSE_NET_STACK_GM",
            accountinglabel="GRO_ENERGY_META",
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetMaxDefense(
                value=Value + NamedReal(name="GRO_ENERGY_META_MAX_DEFENSE_FLAT", value=10 * PLANET_DEFENSE_FACTOR)
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Happiness(low=0),
            stackinggroup="PLANET_SHIELDS_STACK_GEM",
            accountinglabel="GRO_ENERGY_META",
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetMaxShield(
                value=Value + NamedReal(name="GRO_ENERGY_META_MAX_SHIELD_FLAT", value=50 * PLANET_SHIELD_FACTOR)
            ),
        ),
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Focus(type=["FOCUS_INDUSTRY"])
            & Happiness(low=NamedReal(name="GRO_ENERGY_META_MIN_STABILITY", value=20)),
            priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
            effects=SetTargetIndustry(
                value=Value
                + Target.Population
                * NamedReal(name="GRO_ENERGY_META_TARGET_INDUSTRY_PERPOP", value=0.5 * INDUSTRY_PER_POP)
            ),
        ),
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Focus(type=["FOCUS_RESEARCH"])
            & Happiness(low=NamedRealLookup(name="GRO_ENERGY_META_MIN_STABILITY")),
            priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
            effects=SetTargetResearch(
                value=Value
                + Target.Population
                * NamedReal(name="GRO_ENERGY_META_TARGET_RESEARCH_PERPOP", value=1.0 * RESEARCH_PER_POP)
            ),
        ),
    ],
    graphic="icons/tech/pure-energy_metabolism.png",
)



# Autogenerated do not modify manually!
# This is a type-hinting python stub file, used by python IDEs to provide type hints. For more information
# about stub files, see https://www.python.org/dev/peps/pep-0484/#stub-files
# During execution, the actual module is made available via
# a C++ Boost-python process as part of the launch.
from collections.abc import Mapping, Sequence, Set
from enum import IntEnum
from typing import overload

from common.fo_typing import (
    BuildingId,
    BuildingName,
    EmpireId,
    FleetId,
    NamedTuple,
    ObjectId,
    PartName,
    PlanetId,
    PlayerId,
    ShipId,
    SpeciesName,
    SystemId,
    Turn,
)

class AccountingInfo:
    @property
    def causeType(self): ...
    @property
    def customLabel(self): ...
    @property
    def meterChange(self): ...
    @property
    def meterRunningTotal(self): ...
    @property
    def sourceID(self): ...
    @property
    def specificCause(self): ...

class FleetPlan:
    def name(self) -> str: ...
    def ship_designs(self) -> list: ...

class GalaxySetupData:
    @property
    def age(self) -> galaxySetupOptionGeneric: ...
    @property
    def gameUID(self) -> str: ...
    @property
    def maxAIAggression(self) -> aggression: ...
    @property
    def monsterFrequency(self) -> galaxySetupOptionMonsterFreq: ...
    @property
    def nativeFrequency(self) -> galaxySetupOptionGeneric: ...
    @property
    def planetDensity(self) -> galaxySetupOptionGeneric: ...
    @property
    def seed(self) -> str: ...
    @property
    def shape(self) -> galaxyShape: ...
    @property
    def size(self) -> int: ...
    @property
    def specialsFrequency(self) -> galaxySetupOptionGeneric: ...
    @property
    def starlaneFrequency(self) -> galaxySetupOptionGeneric: ...

class GameRules:
    @property
    def empty(self) -> bool: ...
    def getDescription(self, string: str) -> str: ...
    def getDouble(self, string: str) -> float: ...
    def getInt(self, string: str) -> int: ...
    def getRulesAsStrings(self) -> Mapping[str, str]: ...
    def getString(self, string: str) -> str: ...
    def getToggle(self, string: str) -> bool: ...
    def ruleExists(self, string: str) -> bool: ...
    def ruleExistsWithType(self, string: str, rule_type: ruleType) -> bool: ...

class MonsterFleetPlan:
    def locations(self, item_list: list) -> list: ...
    def name(self) -> str: ...
    def ship_designs(self) -> list: ...
    def spawn_limit(self) -> int: ...
    def spawn_rate(self) -> float: ...

class PlayerSetupData:
    @property
    def empire_color(self): ...
    @property
    def empire_name(self): ...
    @property
    def player_name(self): ...
    @property
    def starting_species(self): ...
    @property
    def starting_team(self): ...

class UnlockableItem:
    @property
    def name(self) -> str: ...
    @property
    def type(self) -> unlockableItemType: ...

class buildingType:
    @property
    def costTimeLocationInvariant(self) -> bool: ...
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    def canBeEnqueued(self, number1: int, number2: int) -> bool: ...
    def canBeProduced(self, number1: int, number2: int) -> bool: ...
    def captureResult(self, number1: int, number2: int, number3: int, boolean: bool) -> captureResult: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """
    def perTurnCost(self, number1: int, number2: int) -> float: ...
    def productionCost(self, number1: int, number2: int) -> float: ...
    def productionTime(self, number1: int, number2: int) -> int: ...

class diplomaticMessage(NamedTuple):
    recipient: PlayerId
    sender: PlayerId
    type: diplomaticMessageType

class diplomaticStatusUpdate:
    @property
    def empire1(self) -> EmpireId: ...
    @property
    def empire2(self) -> EmpireId: ...
    @property
    def status(self) -> diplomaticStatus: ...

class empire:
    @property
    def adoptedPolicies(self) -> Sequence[str]: ...
    @property
    def allShipDesigns(self) -> Set[int]: ...
    @property
    def availableBuildingTypes(self) -> Sequence[str]: ...
    @property
    def availablePolicies(self) -> Set[str]: ...
    @property
    def availableShipDesigns(self) -> Set[int]: ...
    @property
    def availableShipHulls(self) -> Sequence[str]: ...
    @property
    def availableShipParts(self) -> Sequence[str]: ...
    @property
    def availableTechs(self) -> Set[str]: ...
    @property
    def capitalID(self) -> PlanetId: ...
    @property
    def categoriesSlotPolicies(self) -> Mapping[str, Mapping[int, str]]: ...
    @property
    def colour(self) -> tuple: ...
    @property
    def eliminated(self) -> bool: ...
    @property
    def empireID(self) -> EmpireId: ...
    @property
    def emptyPolicySlots(self) -> Mapping[str, int]: ...
    @property
    def exploredSystemIDs(self): ...
    @property
    def fleetSupplyableSystemIDs(self) -> Set[int]: ...
    @property
    def lastTurnReceived(self) -> int: ...
    @property
    def name(self) -> str: ...
    @property
    def planetsWithAllocatedPP(self) -> Mapping[Set[int], float]: ...
    @property
    def planetsWithAvailablePP(self) -> Mapping[Set[int], float]: ...
    @property
    def planetsWithWastedPP(self) -> Set[Set[int]]: ...
    @property
    def playerName(self) -> str: ...
    @property
    def productionPoints(self) -> float: ...
    @property
    def productionQueue(self) -> productionQueue: ...
    @property
    def researchQueue(self) -> researchQueue: ...
    @property
    def resourceSupplyGroups(self) -> Set[Set[int]]: ...
    @property
    def supplyUnobstructedSystems(self) -> Set[int]: ...
    @property
    def systemSupplyRanges(self) -> Mapping[int, float]: ...
    @property
    def totalPolicySlots(self) -> Mapping[str, int]: ...
    @property
    def totalShipsOwned(self) -> int: ...
    @property
    def turnsPoliciesAdopted(self) -> Mapping[str, int]: ...
    @property
    def won(self) -> bool: ...
    def buildingTypeAvailable(self, string: str) -> bool: ...
    @overload
    def canBuild(self, build_type: buildType, string: str, number: int) -> bool: ...
    @overload
    def canBuild(self, build_type: buildType, number1: int, number2: int) -> bool: ...
    def canBuild(*args) -> bool: ...
    def getMeter(self, string: str) -> meter:
        """
        Returns the empire meter with the indicated name (string).
        """
    def getTechStatus(self, string: str) -> techStatus: ...
    def hasExploredSystem(self, number: int) -> bool: ...
    def obstructedStarlanes(self) -> Sequence[tuple[int, int]]: ...
    def policyAdopted(self, string: str) -> bool: ...
    def policyAvailable(self, string: str) -> bool: ...
    def policyPrereqsAndExclusionsOK(self, string: str) -> bool: ...
    def population(self) -> float: ...
    def preservedLaneTravel(self, number1: int, number2: int) -> bool: ...
    def productionCostAndTime(self, production_queue_element: productionQueueElement) -> object: ...
    def researchProgress(self, string: str) -> float: ...
    def resourceAvailable(self, resource_type: resourceType) -> float: ...
    def resourceProduction(self, resource_type: resourceType) -> float: ...
    def resourceStockpile(self, resource_type: resourceType) -> float: ...
    def shipDesignAvailable(self, number: int) -> bool: ...
    def slotPolicyAdoptedIn(self, string: str) -> int: ...
    def supplyProjections(self) -> Mapping[SystemId, int]:
        """
        Returns the (negative) number of jumps (int) away each known system ID (int) is from this empire's supply network. 0 in dicates systems that are fleet supplied. -1 indicates a system that is 1 jump away from a supplied system. -4 indicates a system that is 4 jumps from a supply connection.
        """
    def techResearched(self, string: str) -> bool: ...
    def turnPolicyAdopted(self, string: str) -> Turn: ...

class fieldType:
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """

class meter:
    @property
    def current(self) -> float: ...
    @property
    def initial(self) -> float: ...
    def dump(self) -> str:
        """
        Returns string with debug information.
        """

class policy:
    @property
    def category(self) -> str: ...
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def shortDescription(self) -> str: ...
    @overload
    def adoptionCost(self) -> float: ...
    @overload
    def adoptionCost(self, empire_object: empire) -> float: ...
    @overload
    def adoptionCost(self, number: int) -> float: ...
    def adoptionCost(*args) -> float: ...

class productionQueue:
    @property
    def allocatedPP(self) -> Mapping[Set[int], float]: ...
    @property
    def empireID(self) -> EmpireId: ...
    @property
    def empty(self) -> bool: ...
    @property
    def size(self) -> int: ...
    @property
    def totalSpent(self) -> float: ...
    def __getitem__(self, number: int) -> productionQueueElement: ...
    def __iter__(self) -> object: ...
    def __len__(self) -> int: ...

class productionQueueElement:
    @property
    def allocation(self): ...
    @property
    def allowedStockpile(self): ...
    @property
    def blocksize(self): ...
    @property
    def buildType(self): ...
    @property
    def designID(self): ...
    @property
    def locationID(self): ...
    @property
    def name(self) -> str: ...
    @property
    def paused(self): ...
    @property
    def progress(self): ...
    @property
    def remaining(self): ...
    @property
    def removed(self): ...
    @property
    def turnsLeft(self): ...

class researchQueue:
    @property
    def empireID(self) -> EmpireId: ...
    @property
    def empty(self) -> bool: ...
    @property
    def size(self) -> int: ...
    @property
    def totalSpent(self) -> float: ...
    def __contains__(self, research_queue_element: researchQueueElement) -> bool: ...
    def __getitem__(self, number: int) -> researchQueueElement: ...
    def __iter__(self) -> object: ...
    def __len__(self) -> int: ...
    def inQueue(self, string: str) -> bool: ...

class researchQueueElement:
    @property
    def allocation(self): ...
    @property
    def tech(self): ...
    @property
    def turnsLeft(self): ...

class shipDesign:
    @property
    def attack(self) -> float: ...
    @property
    def attackStats(self) -> Sequence[int]: ...
    @property
    def canColonize(self) -> bool: ...
    @property
    def canInvade(self) -> bool: ...
    @property
    def colonyCapacity(self) -> float: ...
    @property
    def costTimeLocationInvariant(self) -> bool: ...
    @property
    def defense(self) -> float: ...
    @property
    def description(self) -> str: ...
    @property
    def designedOnTurn(self) -> Turn: ...
    @property
    def detection(self) -> float: ...
    @property
    def fuel(self) -> float: ...
    @property
    def hasDirectWeapons(self) -> bool: ...
    @property
    def hasFighters(self) -> bool: ...
    @property
    def hull(self) -> str: ...
    @property
    def id(self) -> int: ...
    @property
    def industryGeneration(self) -> float: ...
    @property
    def influenceGeneration(self) -> float: ...
    @property
    def isArmed(self) -> bool: ...
    @property
    def isMonster(self) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def parts(self) -> Sequence[PartName]: ...
    @property
    def researchGeneration(self) -> float: ...
    @property
    def shields(self) -> float: ...
    @property
    def ship_hull(self) -> shipHull: ...
    @property
    def speed(self) -> float: ...
    @property
    def stealth(self) -> float: ...
    @property
    def structure(self) -> float: ...
    @property
    def troopCapacity(self) -> float: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """
    def perTurnCost(self, number1: int, number2: int) -> float: ...
    def productionCost(self, number1: int, number2: int) -> float: ...
    def productionLocationForEmpire(self, number1: int, number2: int) -> bool: ...
    def productionTime(self, number1: int, number2: int) -> int: ...

class shipHull:
    @property
    def costTimeLocationInvariant(self) -> bool: ...
    @property
    def fuel(self) -> float: ...
    @property
    def name(self) -> str: ...
    @property
    def numSlots(self) -> int: ...
    @property
    def slots(self) -> Sequence[shipSlotType]: ...
    @property
    def speed(self) -> float: ...
    @property
    def stealth(self) -> float: ...
    @property
    def structure(self) -> float: ...
    def hasTag(self, string: str) -> bool: ...
    def numSlotsOfSlotType(self, ship_slot_type: shipSlotType) -> int: ...
    def productionCost(self, number1: int, number2: int, number3: int) -> float: ...
    def productionLocation(self, number: int) -> bool:
        """
        Returns the result of Location condition (bool) in passed location_id (int)
        """
    def productionTime(self, number1: int, number2: int, number3: int) -> int: ...

class shipPart:
    @property
    def capacity(self) -> float: ...
    @property
    def costTimeLocationInvariant(self) -> bool: ...
    @property
    def mountableSlotTypes(self) -> Sequence[shipSlotType]: ...
    @property
    def name(self) -> str: ...
    @property
    def partClass(self) -> shipPartClass: ...
    @property
    def secondaryStat(self) -> float: ...
    def canMountInSlotType(self, ship_slot_type: shipSlotType) -> bool: ...
    def hasTag(self, string: str) -> bool: ...
    def productionCost(self, number1: int, number2: int, number3: int) -> float: ...
    def productionLocation(self, number: int) -> bool:
        """
        Returns the result of Location condition (bool) in passed location_id (int)
        """
    def productionTime(self, number1: int, number2: int, number3: int) -> int: ...

class special:
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def spawnlimit(self) -> int: ...
    @property
    def spawnrate(self) -> float: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """
    def initialCapacity(self, number: int) -> float: ...

class species:
    @property
    def canColonize(self) -> bool: ...
    @property
    def canProduceShips(self) -> bool: ...
    @property
    def description(self) -> str: ...
    @property
    def dislikes(self) -> Sequence[str]: ...
    @property
    def foci(self) -> Sequence[str]: ...
    @property
    def homeworlds(self) -> Set[int]: ...
    @property
    def likes(self) -> Sequence[str]: ...
    @property
    def name(self) -> str: ...
    @property
    def native(self) -> bool: ...
    @property
    def preferredFocus(self) -> str: ...
    @property
    def spawnlimit(self) -> int: ...
    @property
    def spawnrate(self) -> float: ...
    @property
    def tags(self) -> Sequence[str]: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """
    def getPlanetEnvironment(self, planet_type: planetType) -> planetEnvironment: ...

class tech:
    @property
    def category(self) -> str: ...
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def prerequisites(self) -> Sequence[str]: ...
    @property
    def shortDescription(self) -> str: ...
    @property
    def unlockedItems(self) -> Sequence[UnlockableItem]: ...
    @property
    def unlockedTechs(self) -> Sequence[str]: ...
    def perTurnCost(self, number: int) -> float: ...
    def recursivePrerequisites(self, number: int) -> Sequence[str]: ...
    def researchCost(self, number: int) -> float: ...
    def researchTime(self, number: int) -> int: ...

class universe:
    @property
    def buildingIDs(self) -> Sequence[BuildingId]: ...
    @property
    def effectAccounting(self) -> Mapping[int, Mapping[meterType, Sequence[AccountingInfo]]]: ...
    @property
    def fieldIDs(self) -> Sequence[int]: ...
    @property
    def fleetIDs(self) -> Sequence[FleetId]: ...
    @property
    def planetIDs(self) -> Sequence[PlanetId]: ...
    @property
    def shipIDs(self) -> Sequence[ShipId]: ...
    @property
    def systemIDs(self) -> Sequence[SystemId]: ...
    def destroyedObjectIDs(self, number: int) -> Set[int]: ...
    def dump(self) -> None: ...
    def getBuilding(self, number: int) -> building: ...
    def getField(self, number: int) -> field: ...
    def getFleet(self, number: int) -> fleet: ...
    def getGenericShipDesign(self, string: str) -> shipDesign:
        """
        Returns the ship design (ShipDesign) with the indicated name (string).
        """
    def getImmediateNeighbors(self, number1: int, number2: int) -> Sequence[int]: ...
    def getObject(self, number: int) -> universeObject: ...
    def getPlanet(self, number: int) -> planet: ...
    def getShip(self, number: int) -> ship: ...
    def getSystem(self, number: int) -> system: ...
    def getSystemNeighborsMap(self, number1: int, number2: int) -> Mapping[int, float]: ...
    def getVisibility(self, number1: int, number2: int) -> visibility: ...
    def getVisibilityTurnsMap(self, number1: int, number2: int) -> Mapping[visibility, int]: ...
    def jumpDistance(self, number1: int, number2: int) -> int:
        """
        If two system ids are passed or both objects are within a system, return the jump distance between the two systems. If one object (e.g. a fleet) is on a starlane, then calculate the jump distance from both ends of the starlane to the target system and return the smaller one.
        """
    def leastJumpsPath(self, number1: int, number2: int, number3: int) -> Sequence[int]: ...
    def linearDistance(self, number1: int, number2: int) -> float: ...
    def shortestNonHostilePath(self, number1: int, number2: int, number3: int) -> Sequence[int]:
        """
        Shortest sequence of System ids and distance from System (number1) to System (number2) with no hostile Fleets as determined by visibility of Empire (number3).  (number3) must be a valid empire.
        """
    def shortestPath(self, number1: int, number2: int, number3: int) -> Sequence[int]: ...
    def shortestPathDistance(self, number1: int, number2: int) -> float: ...
    def statRecords(self) -> Mapping[str, Mapping[int, Mapping[int, float]]]:
        """
        Empire statistics recorded by the server each turn. Indexed first by staistic name (string), then by empire id (int), then by turn number (int), pointing to the statisic value (double).
        """
    def systemHasStarlane(self, number1: int, number2: int) -> bool: ...
    def systemsConnected(self, number1: int, number2: int, number3: int) -> bool: ...
    def updateMeterEstimates(self, obj: object) -> None: ...

class universeObject:
    @property
    def ageInTurns(self): ...
    @property
    def containedObjects(self): ...
    @property
    def containerObject(self): ...
    @property
    def creationTurn(self): ...
    @property
    def id(self) -> ObjectId: ...
    @property
    def meters(self): ...
    @property
    def name(self) -> str: ...
    @property
    def owner(self) -> EmpireId: ...
    @property
    def specials(self) -> SpeciesName: ...
    @property
    def systemID(self) -> SystemId: ...
    @property
    def tags(self): ...
    @property
    def unowned(self): ...
    @property
    def x(self): ...
    @property
    def y(self): ...
    def containedBy(self, number: int) -> bool: ...
    def contains(self, number: int) -> bool: ...
    def currentMeterValue(self, meter_type: meterType) -> float: ...
    def dump(self) -> str:
        """
        Returns string with debug information.
        """
    def getMeter(self, meter_type: meterType) -> meter: ...
    def hasSpecial(self, obj: object) -> bool: ...
    def hasTag(self, string: str) -> bool: ...
    def initialMeterValue(self, meter_type: meterType) -> float: ...
    def ownedBy(self, number: int) -> bool: ...
    def specialAddedOnTurn(self, obj: object) -> int: ...

class building(universeObject):
    @property
    def buildingTypeName(self) -> BuildingName: ...
    @property
    def orderedScrapped(self) -> bool: ...
    @property
    def planetID(self) -> PlanetId: ...
    @property
    def producedByEmpireID(self) -> EmpireId: ...

class field(universeObject):
    @property
    def fieldTypeName(self) -> str: ...
    @overload
    def inField(self, base_object: universeObject) -> bool: ...
    @overload
    def inField(self, floating_number1: float, floating_number2: float) -> bool: ...
    def inField(*args) -> bool: ...

class fleet(universeObject):
    @property
    def aggression(self) -> fleetAggression: ...
    @property
    def aggressive(self) -> bool: ...
    @property
    def canChangeDirectionEnRoute(self) -> bool: ...
    @property
    def empty(self) -> bool: ...
    @property
    def finalDestinationID(self) -> int: ...
    @property
    def fuel(self) -> float: ...
    @property
    def hasArmedShips(self) -> bool: ...
    @property
    def hasColonyShips(self) -> bool: ...
    @property
    def hasFighterShips(self) -> bool: ...
    @property
    def hasMonsters(self) -> bool: ...
    @property
    def hasOutpostShips(self) -> bool: ...
    @property
    def hasTroopShips(self) -> bool: ...
    @property
    def maxFuel(self) -> float: ...
    @property
    def nextSystemID(self) -> int: ...
    @property
    def numShips(self) -> int: ...
    @property
    def obstructive(self) -> bool: ...
    @property
    def previousSystemID(self) -> int: ...
    @property
    def route(self) -> Sequence[int]: ...
    @property
    def shipIDs(self) -> Sequence[ShipId]: ...
    @property
    def speed(self) -> float: ...

class planet(universeObject):
    @property
    def availableFoci(self) -> Sequence[str]: ...
    @property
    def buildingIDs(self) -> Sequence[BuildingId]: ...
    @property
    def clockwiseNextPlanetType(self) -> planetType: ...
    @property
    def counterClockwiseNextPlanetType(self) -> planetType: ...
    @property
    def distanceFromOriginalType(self) -> int: ...
    @property
    def focus(self) -> str: ...
    @property
    def habitableSize(self) -> int: ...
    @property
    def lastColonizedByEmpire(self) -> int: ...
    @property
    def lastInvadedByEmpire(self) -> int: ...
    @property
    def lastTurnAttackedByShip(self) -> int: ...
    @property
    def lastTurnColonized(self) -> int: ...
    @property
    def lastTurnConquered(self) -> int: ...
    @property
    def nextLargerPlanetSize(self) -> planetSize: ...
    @property
    def nextSmallerPlanetSize(self) -> planetSize: ...
    @property
    def originalType(self) -> planetType: ...
    @property
    def ownerBeforeLastConquered(self) -> int: ...
    @property
    def size(self) -> planetSize: ...
    @property
    def speciesName(self) -> SpeciesName: ...
    @property
    def turnsSinceFocusChange(self) -> int: ...
    @property
    def type(self) -> planetType: ...
    def OrbitalPositionOnTurn(self, number: int) -> float: ...
    def environmentForSpecies(self, string: str) -> planetEnvironment: ...
    def nextBetterPlanetTypeForSpecies(self, obj: object, string: str) -> planetType: ...

class ship(universeObject):
    @property
    def arrivedOnTurn(self) -> int: ...
    @property
    def canBombard(self) -> bool: ...
    @property
    def canColonize(self) -> bool: ...
    @property
    def canInvade(self) -> bool: ...
    @property
    def colonyCapacity(self) -> float: ...
    @property
    def design(self) -> shipDesign: ...
    @property
    def designID(self) -> int: ...
    @property
    def fleetID(self) -> int: ...
    @property
    def hasFighters(self) -> bool: ...
    @property
    def isArmed(self) -> bool: ...
    @property
    def isMonster(self) -> bool: ...
    @property
    def lastResuppliedOnTurn(self) -> int: ...
    @property
    def lastTurnActiveInCombat(self) -> int: ...
    @property
    def orderedColonizePlanet(self) -> int: ...
    @property
    def orderedInvadePlanet(self) -> int: ...
    @property
    def orderedScrapped(self) -> bool: ...
    @property
    def partMeters(self) -> Mapping[tuple[meterType, meterType], meter]: ...
    @property
    def producedByEmpireID(self) -> EmpireId: ...
    @property
    def speciesName(self) -> SpeciesName: ...
    @property
    def speed(self) -> float: ...
    @property
    def troopCapacity(self) -> float: ...
    def currentPartMeterValue(self, meter_type: meterType, string: str) -> float: ...
    def initialPartMeterValue(self, meter_type: meterType, string: str) -> float: ...

class system(universeObject):
    @property
    def buildingIDs(self) -> Sequence[BuildingId]: ...
    @property
    def fieldIDs(self) -> Sequence[int]: ...
    @property
    def fleetIDs(self) -> Sequence[FleetId]: ...
    @property
    def lastTurnBattleHere(self) -> int: ...
    @property
    def numStarlanes(self) -> int: ...
    @property
    def planetIDs(self) -> Sequence[PlanetId]: ...
    @property
    def shipIDs(self) -> Sequence[ShipId]: ...
    @property
    def starType(self) -> starType: ...
    @property
    def starlanes(self) -> Sequence[int]: ...
    @property
    def starlanesWormholes(self) -> Sequence[int]: ...
    def HasStarlaneToSystemID(self, number: int) -> bool:
        """
        true if the passed in ID (int) is the ID of a system this system has a starlane connection with
        """

class aggression(IntEnum):
    invalid = -1
    beginner = 0
    turtle = 1
    cautious = 2
    typical = 3
    aggressive = 4
    maniacal = 5

class buildType(IntEnum):
    INVALID_BUILD_TYPE = -1
    BT_NOT_BUILDING = 0
    BT_BUILDING = 1
    BT_SHIP = 2
    BT_PROJECT = 3
    BT_STOCKPILE = 4
    NUM_BUILD_TYPES = 5

class captureResult(IntEnum):
    capture = 0
    destroy = 1
    retain = 2

class diplomaticMessageType(IntEnum):
    noMessage = -1
    warDeclaration = 0
    peaceProposal = 1
    acceptPeaceProposal = 2
    alliesProposal = 3
    acceptAlliesProposal = 4
    endAllies = 5
    cancelProposal = 6
    rejectProposal = 7

class diplomaticStatus(IntEnum):
    war = 0
    peace = 1
    allied = 2

class effectsCauseType(IntEnum):
    invalid = -1
    unknown = 0
    inherent = 1
    tech = 2
    building = 3
    field = 4
    special = 5
    species = 6
    shipPart = 7
    shipHull = 8
    policy = 9

class fleetAggression(IntEnum):
    passive = 0
    defensive = 1
    obstructive = 2
    aggressive = 3

class galaxySetupOptionGeneric(IntEnum):
    invalid = -1
    none = 0
    low = 1
    medium = 2
    high = 3
    random = 4

class galaxySetupOptionMonsterFreq(IntEnum):
    invalid = -1
    none = 0
    extremelyLow = 1
    veryLow = 2
    low = 3
    medium = 4
    high = 5
    veryHigh = 6
    extremelyHigh = 7
    random = 8

class galaxyShape(IntEnum):
    invalid = -1
    spiral2 = 0
    spiral3 = 1
    spiral4 = 2
    cluster = 3
    elliptical = 4
    disc = 5
    box = 6
    irregular = 7
    ring = 8
    random = 9

class meterType(IntEnum):
    targetPopulation = 0
    targetIndustry = 1
    targetResearch = 2
    targetInfluence = 3
    targetConstruction = 4
    targetHappiness = 5
    maxCapacity = 6
    maxSecondaryStat = 7
    maxFuel = 8
    maxShield = 9
    maxStructure = 10
    maxDefense = 11
    maxSupply = 12
    maxStockpile = 13
    maxTroops = 14
    population = 15
    industry = 16
    research = 17
    influence = 18
    construction = 19
    happiness = 20
    capacity = 21
    secondaryStat = 22
    fuel = 23
    shield = 24
    structure = 25
    defense = 26
    supply = 27
    stockpile = 28
    troops = 29
    rebels = 30
    size = 31
    stealth = 32
    detection = 33
    speed = 34

class planetEnvironment(IntEnum):
    uninhabitable = 0
    hostile = 1
    poor = 2
    adequate = 3
    good = 4

class planetSize(IntEnum):
    unknown = -1
    noWorld = 0
    tiny = 1
    small = 2
    medium = 3
    large = 4
    huge = 5
    asteroids = 6
    gasGiant = 7

class planetType(IntEnum):
    unknown = -1
    swamp = 0
    toxic = 1
    inferno = 2
    radiated = 3
    barren = 4
    tundra = 5
    desert = 6
    terran = 7
    ocean = 8
    asteroids = 9
    gasGiant = 10

class resourceType(IntEnum):
    industry = 0
    influence = 1
    research = 2
    stockpile = 3

class roleType(IntEnum):
    host = 0
    clientTypeModerator = 1
    clientTypePlayer = 2
    clientTypeObserver = 3
    galaxySetup = 4

class ruleType(IntEnum):
    invalid = -1
    toggle = 0
    int = 1
    double = 2
    string = 3

class shipPartClass(IntEnum):
    shortRange = 0
    fighterBay = 1
    fighterHangar = 2
    shields = 3
    armour = 4
    troops = 5
    detection = 6
    stealth = 7
    fuel = 8
    colony = 9
    speed = 10
    general = 11
    bombard = 12
    industry = 13
    research = 14
    influence = 15
    productionLocation = 16

class shipSlotType(IntEnum):
    external = 0
    internal = 1
    core = 2

class starType(IntEnum):
    unknown = -1
    blue = 0
    white = 1
    yellow = 2
    orange = 3
    red = 4
    neutron = 5
    blackHole = 6
    noStar = 7

class techStatus(IntEnum):
    unresearchable = 0
    partiallyUnlocked = 1
    researchable = 2
    complete = 3

class unlockableItemType(IntEnum):
    invalid = -1
    building = 0
    shipPart = 1
    shipHull = 2
    shipDesign = 3
    tech = 4
    policy = 5

class visibility(IntEnum):
    invalid = -1
    none = 0
    basic = 1
    partial = 2
    full = 3

def add_special(number: int, string: str) -> None: ...
def all_empires() -> int: ...
def create_building(string: str, number1: int, number2: int) -> int: ...
def create_field(string: str, floating_number1: float, floating_number2: float, floating_number3: float) -> int: ...
def create_field_in_system(string: str, floating_number: float, number: int) -> int: ...
def create_fleet(string: str, number1: int, number2: int, boolean: bool) -> int: ...
def create_monster(string: str, number: int) -> int: ...
def create_monster_fleet(number: int) -> int: ...
def create_planet(planet_size: planetSize, planet_type: planetType, number1: int, number2: int, string: str) -> int: ...
def create_ship(string1: str, string2: str, string3: str, number: int) -> int: ...
def create_system(star_type: starType, string: str, floating_number1: float, floating_number2: float) -> int: ...
def current_turn() -> int: ...
def design_create(
    string1: str,
    string2: str,
    string3: str,
    item_list: list,
    string4: str,
    string5: str,
    boolean: bool,
) -> bool: ...
def design_get_monster_list() -> list: ...
def design_get_premade_list() -> list: ...
def empire_add_ship_design(number: int, string: str) -> None: ...
def empire_set_diplomacy(number1: int, number2: int, arg: diplomaticStatus) -> None: ...
def empire_set_homeworld(number1: int, number2: int, string: str) -> bool: ...
def empire_set_name(number: int, string: str) -> None: ...
def empire_set_stockpile(number: int, resource_type: resourceType, floating_number: float) -> None: ...
def empire_unlock_item(number: int, unlockable_item_type: unlockableItemType, string: str) -> None: ...
@overload
def generate_sitrep(number: int, string1: str, dictionary: dict, string2: str) -> None: ...
@overload
def generate_sitrep(number: int, string1: str, string2: str) -> None: ...
def generate_sitrep(*args) -> None: ...
def generate_starlanes(number1: int, number2: int) -> None: ...
def getBuildingType(string: str) -> buildingType:
    """
    Returns the building type (BuildingType) with the indicated name (string).
    """

def getFieldType(string: str) -> fieldType: ...
def getGameRules() -> GameRules:
    """
    Returns the game rules manager, which can be used to look up the names (string) of rules are defined with what type (boolean / toggle, int, double, string), and what values the rules have in the current game.
    """

def getNamedInt(string: str) -> int:
    """
    Returns the named integer value of the scripted constant with name (string). If no such named constant exists, returns 0.
    """

def getNamedReal(string: str) -> float:
    """
    Returns the named real value of the scripted constant with name (string). If no such named constant exists, returns 0.0.
    """

def getPolicy(string: str) -> policy:
    """
    Returns the policy (Policy) with the indicated name (string).
    """

def getPredefinedShipDesign(string: str) -> shipDesign:
    """
    Returns the ship design (ShipDesign) with the indicated name (string).
    """

def getShipDesign(number: int) -> shipDesign:
    """
    Returns the ship design (ShipDesign) with the indicated id number (int).
    """

def getShipHull(string: str) -> shipHull:
    """
    Returns the ship hull with the indicated name (string).
    """

def getShipPart(string: str) -> shipPart:
    """
    Returns the ShipPart with the indicated name (string).
    """

def getSpecial(string: str) -> special:
    """
    Returns the special (Special) with the indicated name (string).
    """

def getSpecies(string: str) -> species:
    """
    Returns the species (Species) with the indicated name (string).
    """

def getTech(string: str) -> tech:
    """
    Returns the tech (Tech) with the indicated name (string).
    """

def getTechCategories() -> Sequence[str]:
    """
    Returns the names of all tech categories (StringVec).
    """

def get_all_empires() -> list: ...
def get_all_objects() -> list: ...
def get_all_specials() -> list: ...
def get_all_species() -> list: ...
def get_empire(number: int) -> empire: ...
def get_galaxy_setup_data() -> GalaxySetupData: ...
def get_name(number: int) -> object:
    """
    Returns the name (string) of the universe object with the specified object id (int). If there is no such object, returns an empty string and logs an error to the error log.
    """

def get_native_species() -> list: ...
def get_options_db_option_bool(string: str) -> object:
    """
    Returns the bool value of option in OptionsDB or None if the option does not exist.
    """

def get_options_db_option_double(string: str) -> object:
    """
    Returns the double value of option in OptionsDB or None if the option does not exist.
    """

def get_options_db_option_int(string: str) -> object:
    """
    Returns the integer value of option in OptionsDB or None if the option does not exist.
    """

def get_options_db_option_str(string: str) -> object:
    """
    Returns the string value of option in OptionsDB or None if the option does not exist.
    """

def get_owner(number: int) -> int: ...
def get_playable_species() -> list: ...
def get_pos(number: int) -> tuple: ...
def get_resource_dir() -> object: ...
def get_systems() -> list: ...
def get_universe() -> universe: ...
def get_universe_width() -> float: ...
def get_user_config_dir() -> str:
    """
    Returns path to directory where FreeOrion stores user specific configuration.
    """

def get_user_data_dir() -> str:
    """
    Returns path to directory where FreeOrion stores user specific data (saves, etc.).
    """

def get_x(number: int) -> float: ...
def get_y(number: int) -> float: ...
def invalid_object() -> int: ...
def invalid_position() -> float: ...
def jump_distance(number1: int, number2: int) -> int: ...
def large_meter_value() -> float: ...
def linear_distance(number1: int, number2: int) -> float: ...
def load_fleet_plan_list() -> list: ...
def load_monster_fleet_plan_list() -> list: ...
def load_starting_buildings() -> list: ...
def load_unlockable_item_list() -> list: ...
def namedIntDefined(string: str) -> bool:
    """
    Returns true/false (boolean) whether there is a defined int-valued scripted constant with name (string).
    """

def namedRealDefined(string: str) -> bool:
    """
    Returns true/false (boolean) whether there is a defined double-valued scripted constant with name (string).
    """

def objs_get_systems(item_list: list) -> list: ...
def planet_available_foci(number: int) -> list: ...
def planet_cardinal_suffix(number: int) -> object: ...
def planet_get_focus(number: int) -> object: ...
def planet_get_size(number: int) -> planetSize: ...
def planet_get_species(number: int) -> object: ...
def planet_get_type(number: int) -> planetType: ...
def planet_make_colony(number1: int, number2: int, string: str, floating_number: float) -> bool: ...
def planet_make_outpost(number1: int, number2: int) -> bool: ...
def planet_set_focus(number: int, string: str) -> None: ...
def planet_set_size(number: int, planet_size: planetSize) -> None: ...
def planet_set_species(number: int, string: str) -> None: ...
def planet_set_type(number: int, planet_type: planetType) -> None: ...
def policies() -> Sequence[str]:
    """
    Returns the names of all policies (StringVec).
    """

def policiesInCategory(string: str) -> Sequence[str]:
    """
    Returns the names of all policies (StringVec) in the indicated policy category name (string).
    """

def policyCategories() -> Sequence[str]:
    """
    Returns the names of all policy categories (StringVec).
    """

def remove_special(number: int, string: str) -> None: ...
def roman_number(number: int) -> str: ...
def set_name(number: int, string: str) -> None:
    """
    Sets the name (string) of the universe object with the specified object id (int). If there is no such object, just logs an error to the error log.
    """

def set_universe_width(floating_number: float) -> None: ...
def special_has_location(string: str) -> bool: ...
def special_locations(string: str, item_list: list) -> list: ...
def special_spawn_limit(string: str) -> int: ...
def special_spawn_rate(string: str) -> float: ...
def species_add_homeworld(string: str, number: int) -> None: ...
def species_can_colonize(string: str) -> bool: ...
def species_get_planet_environment(string: str, planet_type: planetType) -> planetEnvironment: ...
def species_preferred_focus(string: str) -> object: ...
def species_remove_homeworld(string: str, number: int) -> None: ...
def sys_add_starlane(number1: int, number2: int) -> None: ...
def sys_free_orbits(number: int) -> list: ...
def sys_get_fleets(number: int) -> list: ...
def sys_get_num_orbits(number: int) -> int: ...
def sys_get_planets(number: int) -> list: ...
def sys_get_star_type(number: int) -> starType: ...
def sys_get_starlanes(number: int) -> list: ...
def sys_orbit_occupied(number1: int, number2: int) -> bool: ...
def sys_orbit_of_planet(number1: int, number2: int) -> int: ...
def sys_remove_starlane(number1: int, number2: int) -> None: ...
def sys_set_star_type(number: int, star_type: starType) -> None: ...
def systems_within_jumps_unordered(number: int, item_list: list) -> list:
    """
    Return all systems within ''jumps'' of the systems with ids ''sys_ids''
    """

def techs() -> Sequence[str]:
    """
    Returns the names of all techs (StringVec).
    """

def techsInCategory(string: str) -> Sequence[str]:
    """
    Returns the names of all techs (StringVec) in the indicated tech category name (string).
    """

def userString(string: str) -> str: ...
def userStringExists(string: str) -> bool: ...
def validShipDesign(string: str, item_list: list) -> bool:
    """
    Returns true (boolean) if the passed hull (string) and parts (list of string) make up a valid ship design, and false (boolean) otherwise. Valid ship designs don't have any parts in slots that can't accept that type of part, and contain only hulls and parts that exist (and may also need to contain the correct number of parts - this needs to be verified).
    """



from focs._effects import (
    EffectsGroup,
    Happiness,
    NamedReal,
    OwnedBy,
    Planet,
    SetTargetIndustry,
    Source,
    TargetPopulation,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_AFTER_2ND_SCALING_PRIORITY

Tech(
    name="PRO_ADAPTIVE_AUTOMATION",
    description="PRO_ADAPTIVE_AUTOMATION_DESC",
    short_description="INDUSTRY_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=180 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["LRN_ALGO_ELEGANCE", "PRO_NANOTECH_PROD"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & TargetPopulation(low=0.0001)
            & Happiness(low=NamedReal(name="PRO_ADAPTIVE_AUTO_MIN_STABILITY", value=10)),
            priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
            effects=SetTargetIndustry(value=Value + NamedReal(name="PRO_ADAPTIVE_AUTO_TARGET_INDUSTRY_FLAT", value=2)),
        ),
    ],
    graphic="icons/tech/basic_autofactories.png",
)



from focs._effects import (
    AnyEmpire,
    AsteroidsType,
    Barren,
    Desert,
    EffectsGroup,
    GasGiantType,
    GenerateSitRepMessage,
    Good,
    Inferno,
    IsSource,
    NamedReal,
    Ocean,
    OwnedBy,
    Planet,
    Radiated,
    SetMaxShield,
    SetMaxTroops,
    SetPopulation,
    Source,
    Swamp,
    Target,
    Terran,
    Toxic,
    Tundra,
    Turn,
    Value,
)
from focs._species import *
from macros.misc import PLANET_SHIELD_FACTOR
from macros.priorities import POPULATION_OVERRIDE_PRIORITY
from species.species_macros.empire_opinions import FIXED_OPINION_EFFECTS
from species.species_macros.focus import HAS_ADVANCED_FOCI
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import NO_INDUSTRY
from species.species_macros.influence import NO_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import NO_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import NO_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_ANCIENT_GUARDIANS",
    description="SP_ANCIENT_GUARDIANS_DESC",
    gameplay_description="SP_ANCIENT_GUARDIANS_GAMEPLAY_DESC",
    tags=[
        "ROBOTIC",
        "LOW_BRAINPOWER",
        "NO_RESEARCH",
        "NO_INFLUENCE",
        "AVERAGE_SUPPLY",
        "DESTROYED_ON_CONQUEST",
        "CTRL_STAT_SKIP_DEPOP",
        "PEDIA_ROBOTIC_SPECIES_CLASS",
    ],
    foci=HAS_ADVANCED_FOCI,
    defaultfocus="FOCUS_PROTECTION",
    effectsgroups=[
        # just to make the tags consistent...
        NO_INDUSTRY,
        NO_RESEARCH,
        *NO_INFLUENCE,
        *NO_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        FIXED_OPINION_EFFECTS("SP_ANCIENT_GUARDIANS", 0.0),
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        # not for description
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
        # huge bonus to planetary shields
        EffectsGroup(
            description="ANCIENT_PLANETARY_SHIELD_DESC",
            scope=IsSource & Planet(),
            effects=SetMaxShield(
                value=Value + NamedReal(name="ANCIENT_PLANETARY_SHIELD", value=250 * PLANET_SHIELD_FACTOR)
            ),
        ),
        # huge bonus to ground troops
        EffectsGroup(
            description="ANCIENT_DEFENSE_TROOPS_DESC",
            scope=IsSource & Planet(),
            effects=SetMaxTroops(
                value=Value + (Target.Population * NamedReal(name="ANCIENT_PLANETARY_TROOPS", value=5))
            ),
        ),
        # set population to max
        EffectsGroup(
            scope=IsSource & Planet(),
            activation=Turn(high=1),
            priority=POPULATION_OVERRIDE_PRIORITY,
            effects=SetPopulation(value=Target.TargetPopulation),
        ),
        # self-destruct when captured
        EffectsGroup(
            scope=IsSource,
            activation=Planet() & OwnedBy(affiliation=AnyEmpire),
            priority=POPULATION_OVERRIDE_PRIORITY,
            effects=[
                SetPopulation(value=0),
                GenerateSitRepMessage(
                    message="EFFECT_ANCIENT_GUARDIANS_CAPTURED",
                    label="EFFECT_ANCIENT_GUARDIANS_CAPTURED_LABEL",
                    icon="icons/species/robotic-08.png",
                    parameters={"planet": Source.ID},
                    empire=Source.Owner,
                ),
            ],
        ),
    ],
    environments={
        Swamp: Good,
        Toxic: Good,
        Inferno: Good,
        Radiated: Good,
        Barren: Good,
        Tundra: Good,
        Desert: Good,
        Terran: Good,
        Ocean: Good,
        AsteroidsType: Good,
        GasGiantType: Good,
    },
    graphic="icons/species/robotic-08.png",
    annexation_condition=NoObject,
)



from focs._effects import (
    Capital,
    EffectsGroup,
    IsSource,
    LocalCandidate,
    OwnedBy,
    Planet,
    Population,
    ResourceSupplyConnected,
    SetDetection,
    SetEmpireMeter,
    SetStealth,
    Source,
    Unowned,
    Value,
)
from focs._species import *
from species.species_macros.empire_opinions import FIXED_OPINION_EFFECTS
from species.species_macros.env import VERY_TOLERANT_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.fuel import GREAT_FUEL
from species.species_macros.happiness import GREAT_HAPPINESS
from species.species_macros.industry import GREAT_INDUSTRY
from species.species_macros.influence import GREAT_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.population import GOOD_POPULATION
from species.species_macros.research import GREAT_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import GREAT_STOCKPILE
from species.species_macros.supply import ULTIMATE_SUPPLY
from species.species_macros.troops import GREAT_DEFENSE_TROOPS, GREAT_OFFENSE_TROOPS
from species.species_macros.weapons import ULTIMATE_WEAPONS

Species(
    name="SP_SUPER_TEST",
    description="SP_SUPER_TEST_DESC",
    gameplay_description="SP_SUPER_TEST_GAMEPLAY_DESC",
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "ROBOTIC",
        "LITHIC",
        "ARTISTIC",
        "PHOTOTROPHIC",
        "SELF_SUSTAINING",
        "TELEPATHIC",
        "GREAT_INDUSTRY",
        "GREAT_RESEARCH",
        "GREAT_INFLUENCE",
        "ULTIMATE_WEAPONS",
        "GREAT_FUEL",
        "GOOD_POPULATION",
        "GREAT_HAPPINESS",
        "ULTIMATE_SUPPLY",
        "INFINITE_DETECTION",
        "INFINITE_STEALTH",
        "GREAT_OFFENSE_TROOPS",
        "PEDIA_ARTISTIC",
        "PEDIA_TELEPATHIC_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=["FOCUS_PROTECTION", "SPARK_FOSSILS_SPECIAL", "SUCCULENT_BARNACLES_SPECIAL"],
    dislikes=["FOCUS_GROWTH", "BLD_CULTURE_ARCHIVES", "BLD_CULTURE_LIBRARY"],
    effectsgroups=[
        *GREAT_INDUSTRY,
        *GREAT_RESEARCH,
        *GREAT_INFLUENCE,
        *GREAT_STOCKPILE,
        *GOOD_POPULATION,
        *GREAT_HAPPINESS,
        FIXED_OPINION_EFFECTS("SP_SUPER_TEST", 20),
        *ULTIMATE_SUPPLY,
        *GREAT_DEFENSE_TROOPS,
        *GREAT_OFFENSE_TROOPS,
        *ULTIMATE_WEAPONS,
        *GREAT_FUEL,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
        EffectsGroup(scope=IsSource, effects=SetDetection(value=Value + 10000)),
        EffectsGroup(scope=IsSource, effects=SetStealth(value=Value + 500)),
        EffectsGroup(
            scope=IsSource & Capital,
            effects=SetEmpireMeter(empire=Source.Owner, meter="METER_DETECTION_STRENGTH", value=200.0),
        ),  # Omniscient Detection
    ],
    environments=VERY_TOLERANT_EP,
    annexationcondition=Unowned
    & Population(low=0.001)
    & ResourceSupplyConnected(empire=Source.Owner, condition=Planet() & OwnedBy(empire=Source.Owner)),
    annexationcost=5.0 * LocalCandidate.Population,
    graphic="icons/species/other-04.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SPY_DIST_MOD",
    description="SPY_DIST_MOD_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SPY_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=4,
    tags=["PEDIA_DETECTION_PART_TECHS"],
    prerequisites=["LRN_EVERYTHING"],
    unlock=Item(type=UnlockShipPart, name="SP_DISTORTION_MODULATOR"),
    graphic="icons/ship_parts/distortion_modulator.png",
)



from pytest import mark

from freeorion_tools import get_ai_tag_grade


@mark.parametrize(
    "tags,tags_type",
    (
        (["GREAT_WEAPONS"], "WEAPONS"),
        (["GREAT_WEAPONS"], "weapons"),
        (["GOOD_HULLS", "GREAT_WEAPONS"], "WEAPONS"),
        (["GREAT_WEAPONS", "GOOD_WEAPONS"], "WEAPONS"),
    ),
)
def test_tags_search_by_tag_type_returns_grade(tags, tags_type):
    assert "GREAT" == get_ai_tag_grade(tags, tags_type)


def test_tags_search_by_tag_type_with_underscore_returns_first_grade():
    assert "GREAT" == get_ai_tag_grade(["GREAT_WEAPONS_TEST"], "WEAPONS_TEST")


def test_tags_search_by_not_matching_tag_type_returns_empty_string():
    assert "" == get_ai_tag_grade(["GOOD_HULL"], "WEAPONS")


def test_very_bad_tag_parsing_find_correct_modifier():
    assert "VERY_BAD" == get_ai_tag_grade(["VERY_BAD_WEAPONS"], "WEAPONS")


@mark.parametrize(
    "tags,tags_type",
    (
        (["GREAT_WEAPONS"], "WEAPON"),
        (["GREAT_WEAPONS"], "EAPONS"),
        (["GREAT_WEAPONS"], "EAPON"),
    ),
)
def test_tags_search_by_substring_tag_type_returns_empty_string(tags, tags_type):
    assert "" == get_ai_tag_grade(tags, tags_type)


@mark.parametrize(
    "tags,tags_type",
    (
        (["great_weapons"], "weapons"),
        (["great_weapons"], "WEAPONS"),
    ),
)
def test_lowercase_tags_search_by_tag_type_returns_empty_string(tags, tags_type):
    assert "" == get_ai_tag_grade(tags, tags_type)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_XENOARCH",
    description="LRN_XENOARCH_DESC",
    short_description="XENOARCH_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=70 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_TRANSLING_THT"],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_XENORESURRECTION_LAB"),
        Item(type=UnlockPolicy, name="PLC_DIVINE_AUTHORITY"),
    ],
    graphic="icons/tech/xenoarchaeology.png",
)



from focs._species import *
from species.species_macros.detection import ULTIMATE_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import BARREN_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GOOD_HAPPINESS
from species.species_macros.industry import GOOD_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.population import AVERAGE_POPULATION, LIGHT_SENSITIVE
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stealth import GOOD_STEALTH
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_BANFORO",
    description="SP_BANFORO_DESC",
    gameplay_description="SP_BANFORO_GAMEPLAY_DESC",
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "LITHIC",
        "LIGHT_SENSITIVE",
        "GOOD_INDUSTRY",
        "GOOD_HAPPINESS",
        "AVERAGE_SUPPLY",
        "ULTIMATE_DETECTION",
        "GOOD_STEALTH",
        "CTRL_EXTINCT",
        "PEDIA_LITHIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "SHIMMER_SILK_SPECIAL",
        "FRACTAL_GEODES_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "WORLDTREE_SPECIAL",
        "MONOPOLE_SPECIAL",
        "SUPERCONDUCTOR_SPECIAL",
        "ELERIUM_SPECIAL",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_CONFORMANCE",
        "PLC_ENVIRONMENTALISM",
        "PLC_ALLIED_REPAIR",
        "PLC_INTERSTELLAR_INFRA",
    ],
    dislikes=[
        "BLD_REGIONAL_ADMIN",
        "BLD_GENOME_BANK",
        "BLD_NEUTRONIUM_EXTRACTOR",
        "BLD_SPATIAL_DISTORT_GEN",
        "ECCENTRIC_ORBIT_SPECIAL",
        "KRAKEN_IN_THE_ICE_SPECIAL",
        "KRAKEN_NEST_SPECIAL",
        "RESONANT_MOON_SPECIAL",
        "PLC_CAPITAL_MARKETS",
        "PLC_DIVINE_AUTHORITY",
        "PLC_INDUSTRIALISM",
        "PLC_MARTIAL_LAW",
        "PLC_TERROR_SUPPRESSION",
    ],
    effectsgroups=[
        *GOOD_INDUSTRY,
        *AVERAGE_RESEARCH,
        *AVERAGE_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *GOOD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_BANFORO"),
        *AVERAGE_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        *ULTIMATE_DETECTION,
        *GOOD_STEALTH,
        *LIGHT_SENSITIVE,
        # not for description
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=BARREN_STANDARD_EP,
    graphic="icons/species/banforo.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_MEGA_ECO",
    description="GRO_MEGA_ECO_DESC",
    short_description="GRO_MEGA_ECO_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=18 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=["GRO_PLANET_ECOL"],
    unlock=Item(type=UnlockBuilding, name="BLD_NEST_ERADICATOR"),
    graphic="icons/tech/megafauna_ecology.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_BOMBARD",
    description="SHP_BOMBARD_DESC",
    short_description="SHIP_WEAPON_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=160 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_BOMBARD_WEAPON_TECHS"],
    prerequisites=["SHP_ROOT_AGGRESSION"],
    unlock=Item(type=UnlockPolicy, name="PLC_TERROR_SUPPRESSION"),
    graphic="",
)



from focs._effects import (
    EffectsGroup,
    IsSource,
    SetEmpireMeter,
    Source,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_TIME_MECH",
    description="LRN_TIME_MECH_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_EVERYTHING"],
    unlock=Item(type=UnlockBuilding, name="BLD_STARLANE_NEXUS"),
    effectsgroups=[
        EffectsGroup(
            scope=IsSource,
            effects=SetEmpireMeter(empire=Source.Owner, meter="MILITARY_CATEGORY_NUM_POLICY_SLOTS", value=Value + 1),
        )
    ],
    graphic="icons/tech/temporal_mechanics.png",
)



from focs._effects import (
    Adequate,
    AsteroidsType,
    Barren,
    Desert,
    GasGiantType,
    Inferno,
    Ocean,
    Poor,
    Radiated,
    Swamp,
    Terran,
    Toxic,
    Tundra,
    Uninhabitable,
)
from focs._species import *
from species.species_macros.empire_opinions import FIXED_OPINION_EFFECTS
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import GREAT_HAPPINESS
from species.species_macros.industry import VERY_BAD_INDUSTRY
from species.species_macros.influence import NO_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import VERY_BAD_RESEARCH
from species.species_macros.shields import BAD_SHIP_SHIELDS
from species.species_macros.stockpile import NO_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS
from species.species_macros.weapons import BAD_WEAPONS

Species(
    name="SP_EXOBOT",
    description="SP_EXOBOT_DESC",
    gameplay_description="SP_EXOBOT_GAMEPLAY_DESC",
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ROBOTIC",
        "LOW_BRAINPOWER",
        "VERY_BAD_INDUSTRY",
        "VERY_BAD_RESEARCH",
        "NO_INFLUENCE",
        "BAD_POPULATION",
        "BAD_WEAPONS",
        "BAD_SHIP_SHIELDS",
        "GREAT_HAPPINESS",
        "BAD_SUPPLY",
        "CTRL_ALWAYS_REPORT",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_ROBOTIC_SPECIES_CLASS",
        "NO_TERRAFORM",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        # [[HAS_INFLUENCE_FOCUS]]
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "PLC_ALLIED_REPAIR",
        "PLC_BUREAUCRACY",
        "PLC_DESIGN_SIMPLICITY",
        "PLC_ENGINEERING",
        "PLC_EXOBOT_PRODUCTIVITY",
    ],
    dislikes=[
        "BLD_BLACK_HOLE_POW_GEN",
        "BLD_INDUSTRY_CENTER",
        "BLD_LIGHTHOUSE",
        "BLD_SCANNING_FACILITY",
        "BLD_SHIPYARD_ENRG_COMP",
        "BLD_SOL_ORB_GEN",
    ],
    effectsgroups=[
        *VERY_BAD_INDUSTRY,
        *VERY_BAD_RESEARCH,
        *NO_INFLUENCE,
        *NO_STOCKPILE,
        *BAD_POPULATION,
        *GREAT_HAPPINESS,
        FIXED_OPINION_EFFECTS("SP_EXOBOT", 20),
        *BAD_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        *BAD_WEAPONS,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *BAD_SHIP_SHIELDS,
    ],
    environments={
        Swamp: Adequate,
        Toxic: Adequate,
        Inferno: Adequate,
        Radiated: Adequate,
        Barren: Adequate,
        Tundra: Adequate,
        Desert: Adequate,
        Terran: Adequate,
        Ocean: Adequate,
        AsteroidsType: Poor,
        GasGiantType: Uninhabitable,
    },
    graphic="icons/species/robotic-01.png",
)



"""
This module stores information about game rules and content that can not be dynamically accessed.

If other modules rely on information regarding the game content that can not be queried from the server and thus
 needs to be hardcoded, the information should be implemented here and referenced by the other module.

The game content is sorted by categories:
    - Basic Game Rules
    - Species
    - Specials
    - Techs
    - Buildings
    - Shipdesign / -parts
Each category is wrapped by <editor-fold desc="CATEGORY">   </editor-fold> tags which allow code-folding in IDE.
Categories may have subcategories which are also wrapped by these tags. When adding new content, try to keep it
structured in the aforementioned way. Use the category names to find the appropriate section of the file.

Example usage:
    import AIDependencies
    my_industry = AIDependencies.INDUSTRY_PER_POP * my_population
"""
import freeOrionAIInterface as fo

# Note re common dictionary lookup structure, "PlanetSize-Dependent-Lookup":
# Many dictionaries herein (a prime example being the building_supply dictionary) have a primary key (such as
# special_name, or building_type_name), and then provide a sub-dictionary with which to look up the resultant value
# which may in some cases depend upon PlanetSize; if not then the key -1 stands for any planet size.
# So, if the value to be provided by the sub-dictionary does not depend on PlanetSize, then this sub-dictionary
# should be of the form
#    {-1: return_val}
# if the return value *is* dependent on PlanetSize, then the sub-dictionary should be of the form:
#    {fo.planetSize.tiny: return_val_1,
#     fo.planetSize.small: return_val_2,
#     fo.planetSize.medium: return_val_3,
#     fo.planetSize.large: return_val_4,
#     fo.planetSize.huge: return_val_5,
#     fo.planetSize.gasGiant: return_val_6,
#     }
# Please refer to the building_supply dictionary below for an example that displays both of these uses.

# <editor-fold desc="Basic Game Rules">
# TODO look into (enabling) simply retrieving the below values via UserString
INDUSTRY_PER_POP = 0.2
RESEARCH_PER_POP = 0.2
TROOPS_PER_POP = 0.2

PROT_FOCUS_MULTIPLIER = 2.0
HOMEWORLD_INFLUENCE_COST = 1  # homeworld independency movement

STABILITY_PER_LIKED_FOCUS = 2.0
STABILITY_HOMEWORLD_BONUS = 5.0
STABILITY_PER_LIKED_BUILDING_ON_PLANET = 4.0
STABILITY_PER_LIKED_BUILDING_IN_SYSTEM = 1.0
STABILITY_BASE_LIKED_BUILDING_ELSEWHERE = 0.5  # bonus is this time sqrt(number)
STABILITY_BY_WORLDTREE = 1.0
# specials (dis)likes are not affected by PlanetUtilsAI.dislike_factor() and also not by who owns them!
STABILITY_PER_LIKED_SPECIAL_ON_PLANET = 3.0
STABILITY_PER_LIKED_SPECIAL_IN_SYSTEM = 1.0


SHIP_STRUCTURE_FACTOR = fo.getGameRules().getDouble("RULE_SHIP_STRUCTURE_FACTOR")
SHIP_WEAPON_DAMAGE_FACTOR = fo.getGameRules().getDouble("RULE_SHIP_WEAPON_DAMAGE_FACTOR")
FIGHTER_DAMAGE_FACTOR = fo.getGameRules().getDouble("RULE_FIGHTER_DAMAGE_FACTOR")

# Colonization details
OUTPOST_POD_COST = 50  # TODO Query directly from part
COLONY_POD_UPKEEP = 0.06
SHIP_UPKEEP = 0.01

#  Supply details
supply_by_size = {
    fo.planetSize.tiny: 2,
    fo.planetSize.small: 1,
    fo.planetSize.large: -1,
    fo.planetSize.huge: -2,
    fo.planetSize.gasGiant: -1,
}

# Drydock repair details
DRYDOCK_HAPPINESS_THRESHOLD = 5

# Constants defined by the C++ game engine
INVALID_ID = -1
ALL_EMPIRES = -1

# Monster fleet plans
MINIMUM_GUARD_DISTANCE_TO_HOME_SYSTEM = 2


class Tags:
    POPULATION = "POPULATION"
    INDUSTRY = "INDUSTRY"
    INFLUENCE = "INFLUENCE"
    WEAPONS = "WEAPONS"
    RESEARCH = "RESEARCH"
    SUPPLY = "SUPPLY"
    STABILITY = "HAPPINESS"
    ATTACKTROOPS = "OFFENSE_TROOPS"
    STEALTH = "STEALTH"
    SHIP_SHIELDS = "SHIP_SHIELDS"
    FUEL = "FUEL"
    DETECTION = "DETECTION"
    INDEPENDENT = "INDEPENDENT_HAPPINESS"
    ARTISTIC = "ARTISTIC"
    XENOPHOBIC = "XENOPHOBIC"


# </editor-fold>


# <editor-fold desc="Environmental population modifiers">
#  Population modifiers that are based on environment

# Population modifiers scaling with planet size
#  - [[EARLY_PRIORITY]]: Affected by species modifiers
POP_SIZE_MOD_MAP_MODIFIED_BY_SPECIES = {
    # "Effect": [uninhab, hostile, poor, adequate, good]
    "environment_bonus": [0, -4, -2, 0, 3],
    "GRO_SYMBIOTIC_BIO": [0, 0, 1, 1, 1],
    "GRO_XENO_GENETICS": [0, 1, 2, 2, 0],
    "GRO_XENO_HYBRIDS": [0, 2, 1, 0, 0],
}

# Population modifiers scaling with planet size
#  - [[LATE_PRIORITY]] or later:
POP_SIZE_MOD_MAP_NOT_MODIFIED_BY_SPECIES = {
    "GRO_SUBTER_HAB": [0, 1, 1, 1, 1],
    "CON_NDIM_STRC": [0, 2, 2, 2, 2],
    "CON_ORBITAL_HAB": [0, 1, 1, 1, 1],
}

# flat boni not dependent on size
POP_CONST_MOD_MAP = {
    # "Source": [uninhab, hostile, poor, adequate, good]
    "GRO_PLANET_ECOL": [0, 0, 0, 1, 1],
    "GRO_SYMBIOTIC_BIO": [0, 0, 0, -1, -1],
}

# phototrophic star coefficients
POP_MOD_PHOTOTROPHIC_STAR_MAP = {
    fo.starType.blue: 3,
    fo.starType.white: 1.5,
    fo.starType.red: -1,
    fo.starType.neutron: -1,
    fo.starType.blackHole: -10,
    fo.starType.noStar: -10,
}

# lightsensitive star coefficients
TAG_LIGHT_SENSITIVE = "LIGHT_SENSITIVE"
POP_MOD_LIGHTSENSITIVE_STAR_MAP = {fo.starType.blue: -2, fo.starType.white: -1}

# the percentage of normal population that a species with Gaseous tag has
GASEOUS_POP_FACTOR = 0.50
# </editor-fold>

# <editor-fold desc="Detection Strengths">
DETECTION_TECH_STRENGTHS = {
    "SPY_DETECT_1": 10,
    "SPY_DETECT_2": 30,
    "SPY_DETECT_3": 50,
    "SPY_DETECT_4": 70,
    "SPY_DETECT_5": 200,
}
# </editor-fold>

# <editor-fold desc="Stealth Strengths">
PRIMARY_FOCS_STEALTH_LEVELS = {
    "LOW_STEALTH": 20,
    "MEDIUM_STEALTH": 40,
    "HIGH_STEALTH": 60,
    "VERY_HIGH_STEALTH": 80,
}

STEALTH_SPECIAL_STRENGTHS = {
    "CLOUD_COVER_SLAVE_SPECIAL": PRIMARY_FOCS_STEALTH_LEVELS["LOW_STEALTH"],
    "VOLCANIC_ASH_SLAVE_SPECIAL": PRIMARY_FOCS_STEALTH_LEVELS["MEDIUM_STEALTH"],
    "DIM_RIFT_SLAVE_SPECIAL": PRIMARY_FOCS_STEALTH_LEVELS["HIGH_STEALTH"],
    "VOID_SLAVE_SPECIAL": PRIMARY_FOCS_STEALTH_LEVELS["VERY_HIGH_STEALTH"],
}

BASE_PLANET_STEALTH = 5

STEALTH_STRENGTHS_BY_SPECIES_TAG = {
    "BAD": -PRIMARY_FOCS_STEALTH_LEVELS["LOW_STEALTH"],
    "GOOD": PRIMARY_FOCS_STEALTH_LEVELS["LOW_STEALTH"],
    "GREAT": PRIMARY_FOCS_STEALTH_LEVELS["MEDIUM_STEALTH"],
    "ULTIMATE": PRIMARY_FOCS_STEALTH_LEVELS["HIGH_STEALTH"],
    "INFINITE": 500,  # used by super testers
}
# </editor-fold>

# <editor-fold desc="Specials">
# <editor-fold desc="Growth Focus specials">
# stores growth special per metabolism
metabolismBoostMap = {
    "ORGANIC": [
        "FRUIT_SPECIAL",
        "PROBIOTIC_SPECIAL",
        "SPICE_SPECIAL",
    ],
    "LITHIC": [
        "CRYSTALS_SPECIAL",
        "ELERIUM_SPECIAL",
        "MINERALS_SPECIAL",
    ],
    "ROBOTIC": [
        "MONOPOLE_SPECIAL",
        "POSITRONIUM_SPECIAL",
        "SUPERCONDUCTOR_SPECIAL",
    ],
    "SELF_SUSTAINING": [],
}

# stores which metabolism profits from each special - autogenerated DO NOT EDIT
metabolismBoosts = {}
for metab, boosts in metabolismBoostMap.items():
    for boost in boosts:
        metabolismBoosts[boost] = metab
# </editor-fold>

# <editor-fold desc="Industry boosting specials">
# Each adds INDUSTRY_PER_POP before production is multiplied by species skill modifier
industry_boost_specials_modified = {
    "TIDAL_LOCK_SPECIAL",
}
# Each adds INDUSTRY_PER_POP after all multipliers have been applied
industry_boost_specials_unmodified = {
    "CRYSTALS_SPECIAL",
    "ELERIUM_SPECIAL",
    "MINERALS_SPECIAL",
    "MONOPOLE_SPECIAL",
    "POSITRONIUM_SPECIAL",
    "SUPERCONDUCTOR_SPECIAL",
}
# </editor-fold>

luxury_specials = {
    "FRACTAL_GEODES_SPECIAL",
    "MIMETIC_ALLOY_SPECIAL",
    "SHIMMER_SILK_SPECIAL",
    "SPARK_FOSSILS_SPECIAL",
    "SUCCULENT_BARNACLES_SPECIAL",
}

# <editor-fold desc="Other Population changing specials">
# Please see the Note at top of this file regarding PlanetSize-Dependent-Lookup
# Regardless of whether the sub-dictionary here has PlanetSize keys, the final
# value will be applied as a *fixed-size mod* to the max population
WORLDTREE_SPECIAL = "WORLDTREE_SPECIAL"
POP_FIXED_MOD_SPECIALS = {
    WORLDTREE_SPECIAL: 1,  # Not for SP_KHAKTURIAN...
}
GAIA_SPECIAL = "GAIA_SPECIAL"


def not_affect_by_special(special: str, species: str) -> bool:
    if special == WORLDTREE_SPECIAL:
        return species == "SP_KHAKTURIAN"
    # They do not have a good environment on normal planets (or none at all).
    # TBD: could determine that list when there is a way to iterate over all species
    if special == GAIA_SPECIAL:
        return species in ("SP_EXOBOT", "SP_SLY", "SP_THENIAN")


# Please see the Note at top of this file regarding PlanetSize-Dependent-Lookup
# The return value from the respective sub-dictionary will be applied as a
# population modifier proportional to PlanetSize, i.e.,
#       max_pop += return_val * PlanetSize
# So, most commonly the lookup key for the sub-dictionary here will be
# PlanetSize independent (i.e., -1), although it is possible to use
# PlanetSize keys here for a more complex interaction.
# Regardless of whether the sub-dictionary here has PlanetSize keys, the final
# value will be applied as a fixed-size mod to the max population
POP_PROPORTIONAL_MOD_SPECIALS = {
    "TIDAL_LOCK_SPECIAL": -1,
    "TEMPORAL_ANOMALY_SPECIAL": -5,
}
# </editor-fold>

# <editor-fold desc="Industry related specials">
HONEYCOMB_SPECIAL = "HONEYCOMB_SPECIAL"
# </editor-fold>

# <editor-fold desc="Research related specials">
COMPUTRONIUM_SPECIAL = "COMPUTRONIUM_SPECIAL"

ANCIENT_RUINS_SPECIAL = "ANCIENT_RUINS_SPECIAL"
ANCIENT_RUINS_SPECIAL2 = "ANCIENT_RUINS_DEPLETED_SPECIAL"  # nothing to excavate anymore, but some research bonus
ASTEROID_COATING_OWNED_SPECIAL = "ASTEROID_COATING_OWNED_SPECIAL"
ASTEROID_COATING_SPECIAL = "ASTEROID_COATING_SPECIAL"
# </editor-fold>

# <editor-fold desc="Supply related specials">
SUPPLY_MOD_SPECIALS = {
    WORLDTREE_SPECIAL: {
        -1: 1,
    },
    "ECCENTRIC_ORBIT_SPECIAL": {
        -1: -2,
    },
    "ACCRETION_DISC_SPECIAL": {
        -1: -1,
    },
}
# </editor-fold>

# <editor-fold desc="Native Tech/Defense related specials">
# (special_name, defense&shield_bonus)
TECH_NATIVE_SPECIALS = {
    "MODERATE_TECH_NATIVES_SPECIAL": {"defense": 10, "shield": 10},
    "HIGH_TECH_NATIVES_SPECIAL": {"defense": 30, "shield": 30},
}
# </editor-fold>


# </editor-fold>


# <editor-fold desc="Techs">
# <editor-fold desc="Various tech names">
# For IDE support and to avoid typos and easier editing, define tech names here and import for scripts
GRO_LIFE_CYCLE = "GRO_LIFECYCLE_MAN"
PRO_ORBITAL_GEN = "PRO_ORBITAL_GEN"
PRO_SOL_ORB_GEN = "PRO_SOL_ORB_GEN"
PRO_MICROGRAV_MAN = "PRO_MICROGRAV_MAN"
PRO_SINGULAR_GEN = "PRO_SINGULAR_GEN"
PRO_AUTO_1 = "PRO_ADAPTIVE_AUTOMATION"
PRO_AUTO_2 = "PRO_SENTIENT_AUTOMATION"
PRO_NEUTRONIUM_EXTRACTION = "PRO_NEUTRONIUM_EXTRACTION"
NEST_DOMESTICATION_TECH = "SHP_DOMESTIC_MONSTER"

LRN_ARTIF_MINDS_1 = "LRN_NASCENT_AI"
LRN_ALGO_ELEGANCE = "LRN_ALGO_ELEGANCE"
GRO_PLANET_ECOL = "GRO_PLANET_ECOL"
LRN_QUANT_NET = "LRN_QUANT_NET"
LRN_XENOARCH = "LRN_XENOARCH"
LRN_EVERYTHING = "LRN_EVERYTHING"
LRN_ART_BLACK_HOLE = "LRN_ART_BLACK_HOLE"

SR_FLUX_LANCE = "SR_FLUX_LANCE"
SPY_STEALTH_1 = "SPY_STEALTH_1"
SPY_STEALTH_2 = "SPY_STEALTH_2"

EXOBOT_TECH_NAME = "PRO_EXOBOTS"
SHP_ASTEROID_HULLS = "SHP_ASTEROID_HULLS"

# </editor-fold>


# Tech required to build outpost pods
OUTPOSTING_TECH = "SHP_GAL_EXPLO"

# <editor-fold desc="Defense techs">
# TODO are these 2 necessary? Could rely on ordered DEFENSE_SHIELD_TECHS (Morlic)
PLANET_BARRIER_I_TECH = "DEF_PLAN_BARRIER_SHLD_1"
DEFENSE_REGEN_1_TECH = "DEF_DEFENSE_NET_REGEN_1"
# </editor-fold>

# <editor-fold desc="Fuel techs">
FUEL_TANK_UPGRADE_DICT = {
    # "PARTNAME": tuple((tech_name, fuel_upgrade), (tech_name2, fuel_upgrade2), ...)
    "FU_BASIC_TANK": (("SHP_DEUTERIUM_TANK", 0.5), ("SHP_ANTIMATTER_TANK", 1.5)),
    "FU_RAMSCOOP": (),
    "FU_ZERO_FUEL": (),
}

# DO NOT TOUCH THIS ENTRY BUT UPDATE FUEL_TANK_UPGRADE_DICT INSTEAD!
FUEL_UPGRADE_TECHS = frozenset(
    tech_name for _dict in (FUEL_TANK_UPGRADE_DICT,) for tups in _dict.values() for (tech_name, _) in tups
)
# </editor-fold>

# <editor-fold desc="Weapon techs">
# TODO (Morlic): Consider using only 1 dict with (capacity, secondaryStat) tuple as entries
WEAPON_UPGRADE_DICT = {
    part_name: tuple(
        {tech_name: damage * SHIP_WEAPON_DAMAGE_FACTOR for tech_name, damage, in techs_damage.items()}.items()
    )
    for part_name, techs_damage in {
        # Output "PARTNAME": tuple((tech_name, dmg_upgrade), (tech_name2, dmg_upgrade2), ...)
        #  Input "PARTNAME": {tech_name: dmg_upgrade, tech_name2, dmg_upgrade2, ...}
        "SR_WEAPON_0_1": {},
        "SR_WEAPON_1_1": {"SHP_WEAPON_1_%d" % i: 1 for i in [2, 3, 4]},
        "SR_WEAPON_2_1": {"SHP_WEAPON_2_%d" % i: 2 for i in [2, 3, 4]},
        "SR_WEAPON_3_1": {"SHP_WEAPON_3_%d" % i: 3 for i in [2, 3, 4]},
        "SR_WEAPON_4_1": {"SHP_WEAPON_4_%d" % i: 5 for i in [2, 3, 4]},
        "SR_ARC_DISRUPTOR": {"SHP_WEAPON_ARC_DISRUPTOR_%d" % i: i for i in [2, 3]},
        SR_FLUX_LANCE: {},
        "SR_SPINAL_ANTIMATTER": {},
        "SR_GRAV_PULSE": {},
    }.items()
}

# ROF == rate of fire == weapon shots
WEAPON_ROF_UPGRADE_DICT = {
    # "PARTNAME": tuple((tech_name, rof_upgrade), (tech_name2, rof_upgrade2), ...)
    "SR_WEAPON_0_1": (("SHP_WEAPON_1_3", 1), ("SHP_WEAPON_2_3", 1), ("SHP_WEAPON_3_3", 1), ("SHP_WEAPON_4_3", 1)),
    "SR_WEAPON_1_1": (),
    "SR_WEAPON_2_1": (),
    "SR_WEAPON_3_1": (),
    "SR_WEAPON_4_1": (),
    "SR_ARC_DISRUPTOR": (),
    SR_FLUX_LANCE: (),
    "SR_SPINAL_ANTIMATTER": (),
    "SR_GRAV_PULSE": (),
}

FIGHTER_DAMAGE_UPGRADE_DICT = {
    part_name: tuple({tech_name: damage * FIGHTER_DAMAGE_FACTOR for tech_name, damage, in techs_damage.items()}.items())
    for part_name, techs_damage in {
        # Output "PARTNAME": tuple((tech_name, dmg_upgrade), (tech_name2, dmg_upgrade2), ...)
        #  Input "PARTNAME": {tech_name: dmg_upgrade, tech_name2: dmg_upgrade2, ...}
        "FT_HANGAR_1": {"SHP_FIGHTERS_2": 0, "SHP_FIGHTERS_3": 0, "SHP_FIGHTERS_4": 0},
        "FT_HANGAR_2": {"SHP_FIGHTERS_2": 2, "SHP_FIGHTERS_3": 2, "SHP_FIGHTERS_4": 2},
        "FT_HANGAR_3": {"SHP_FIGHTERS_2": 3, "SHP_FIGHTERS_3": 3, "SHP_FIGHTERS_4": 3},
        "FT_HANGAR_4": {"SHP_FIGHTERS_2": 6, "SHP_FIGHTERS_3": 6, "SHP_FIGHTERS_4": 6},
    }.items()
}

FIGHTER_CAPACITY_UPGRADE_DICT = {
    # "PARTNAME": tuple((tech_name, capacity_upgrade), (tech_name2, capacity_upgrade2), ...)
    "FT_HANGAR_1": (("SHP_FIGHTERS_2", 1), ("SHP_FIGHTERS_3", 1), ("SHP_FIGHTERS_4", 1)),
    "FT_HANGAR_2": (),
    "FT_HANGAR_3": (),
    "FT_HANGAR_4": (),
}

# DO NOT TOUCH THIS ENTRY BUT UPDATE WEAPON_UPGRADE_DICT INSTEAD!
WEAPON_UPGRADE_TECHS = frozenset(
    tech_name
    for _dict in (WEAPON_UPGRADE_DICT, WEAPON_ROF_UPGRADE_DICT)
    for tups in _dict.values()
    for (tech_name, _) in tups
)
FIGHTER_UPGRADE_TECHS = frozenset(
    tech_name
    for _dict in (FIGHTER_DAMAGE_UPGRADE_DICT, FIGHTER_CAPACITY_UPGRADE_DICT)
    for tups in _dict.values()
    for (tech_name, _) in tups
)
# </editor-fold>

# <editor-fold desc="Tech Groups for Automated Research Approach">
# TODO obtain this information from techs
UNRESEARCHABLE_TECHS = (
    "SHP_KRILL_SPAWN",
    "DEF_PLANET_CLOAK",
    "TECH_COL_BANFORO",
    "TECH_COL_KILANDOW",
    "TECH_COL_MISIORLA",
    "SPY_DETECT_1",
    "SPY_PLANET_STEALTH_MOD",
)
# </editor-fold>

# </editor-fold>


# <editor-fold desc="Species">
# species names
SP_LEMBALALAM = "SP_LEMBALALAM"

# Species modifiers
SPECIES_RESEARCH_MODIFIER = {"NO": 0.0, "VERY_BAD": 0.5, "BAD": 0.75, "GOOD": 1.5, "GREAT": 2.0, "ULTIMATE": 3.0}
SPECIES_INDUSTRY_MODIFIER = {"NO": 0.0, "VERY_BAD": 0.5, "BAD": 0.75, "GOOD": 1.5, "GREAT": 2.0, "ULTIMATE": 3.0}
SPECIES_INFLUENCE_MODIFIER = {"NO": 0.0, "VERY_BAD": 0.5, "BAD": 0.75, "GOOD": 1.5, "GREAT": 2.0, "ULTIMATE": 3.0}
SPECIES_POPULATION_MODIFIER = {"EXTREMELY_BAD": 0.25, "VERY_BAD": 0.5, "BAD": 0.75, "GOOD": 1.25}
SPECIES_SUPPLY_MODIFIER = {"VERY_BAD": -1, "BAD": 0, "AVERAGE": 1, "GREAT": 2, "ULTIMATE": 3}
SPECIES_STABILITY_MODIFIER = {"VERY_BAD": -5.0, "BAD": -2.5, "AVERAGE": 0, "GOOD": 2.5, "GREAT": 5.0, "ULTIMATE": 7.5}
SPECIES_FUEL_MODIFIER = {"NO": -100, "BAD": -0.5, "AVERAGE": 0, "GOOD": 0.5, "GREAT": 1, "ULTIMATE": 1.5}
# the actual effect depends on the type of shield, see CombatRatingsAI.species_shield_bonus
SPECIES_SHIP_SHIELD_MODIFIER = {"BAD": -0.5, "AVERAGE": 0, "GOOD": 0.5, "GREAT": 1, "ULTIMATE": 1.5}
# values for both offense and defense
SPECIES_TROOP_MODIFIER = {"NO": 0.0, "BAD": 0.5, "": 1.0, "GOOD": 1.5, "GREAT": 2.0, "ULTIMATE": 3.0}
# missing: HAPPINESS, ...

# <editor-fold desc="XenoResurrectionSpecies">
EXTINCT_SPECIES = ["BANFORO", "KILANDOW", "MISIORLA"]
# </editor-fold>

# <editor-fold desc="Piloting traits">
# TODO (Morlic): Consider using only 1 dict with tuple for (Capacity, SecondaryStat) effects
PILOT_DAMAGE_UNSCALED_MODIFIER_DICT = {
    # TRAIT:    {weapon_name: effect, weapon_name2: effect2,...}
    "NO": {},
    "BAD": {"SR_WEAPON_1_1": -1, "SR_WEAPON_2_1": -2, "SR_WEAPON_3_1": -3, "SR_WEAPON_4_1": -5},
    "GOOD": {"SR_WEAPON_1_1": 1, "SR_WEAPON_2_1": 2, "SR_WEAPON_3_1": 3, "SR_WEAPON_4_1": 5},
    "GREAT": {"SR_WEAPON_1_1": 2, "SR_WEAPON_2_1": 4, "SR_WEAPON_3_1": 6, "SR_WEAPON_4_1": 10},
    "ULTIMATE": {"SR_WEAPON_1_1": 3, "SR_WEAPON_2_1": 6, "SR_WEAPON_3_1": 9, "SR_WEAPON_4_1": 15},
}

PILOT_ROF_MODIFIER_DICT = {
    # TRAIT:    {weapon_name: effect, weapon_name2: effect2,...}
    "NO": {},
    "BAD": {"SR_WEAPON_0_1": -1},
    "GOOD": {"SR_WEAPON_0_1": 1},
    "GREAT": {"SR_WEAPON_0_1": 2},
    "ULTIMATE": {"SR_WEAPON_0_1": 3},
}

PILOT_FIGHTERDAMAGE_UNSCALED_MODIFIER_DICT = {
    # TRAIT:    {hangar_name: effect, hangar_name2: effect2,...}
    # Note: FT_HANGAR_1 fighters are not able to attack ships so pilot damage modifier does not apply
    "NO": {},
    "BAD": {"FT_HANGAR_1": 0, "FT_HANGAR_2": -1, "FT_HANGAR_3": -1, "FT_HANGAR_4": -1},
    "GOOD": {"FT_HANGAR_1": 0, "FT_HANGAR_2": 1.5, "FT_HANGAR_3": 1.5, "FT_HANGAR_4": 1.5},
    "GREAT": {"FT_HANGAR_1": 0, "FT_HANGAR_2": 3, "FT_HANGAR_3": 3, "FT_HANGAR_4": 3},
    "ULTIMATE": {"FT_HANGAR_1": 0, "FT_HANGAR_2": 4.5, "FT_HANGAR_3": 4.5, "FT_HANGAR_4": 4.5},
}

PILOT_FIGHTER_CAPACITY_MODIFIER_DICT = {
    # TRAIT:    {hangar_name: effect, hangar_name2: effect2,...}
    "NO": {},
    "BAD": {},
    "GOOD": {},
    "GREAT": {},
    "ULTIMATE": {},
}

HANGAR_LAUNCH_CAPACITY_MODIFIER_DICT = {
    # hangar_name: {bay_name: ((tech_name, effect), ...), bay_name2: ((tech_name, effect), ...}
    "FT_HANGAR_1": {
        "FT_BAY_1": (("SHP_FIGHTERS_1", 1), ("SHP_FIGHTERS_2", 1), ("SHP_FIGHTERS_3", 1), ("SHP_FIGHTERS_4", 1))
    },
}


def _scale_part_damage(part_damage: dict[str, int], factor: float) -> dict[str, float]:
    scaled_part_damage = {weapon_name: (damage * factor) for weapon_name, damage in part_damage.items()}
    return scaled_part_damage


PILOT_DAMAGE_MODIFIER_DICT = {
    trait: _scale_part_damage(part_damage, SHIP_WEAPON_DAMAGE_FACTOR)
    for trait, part_damage in PILOT_DAMAGE_UNSCALED_MODIFIER_DICT.items()
}

PILOT_FIGHTERDAMAGE_MODIFIER_DICT = {
    trait: _scale_part_damage(part_damage, FIGHTER_DAMAGE_FACTOR)
    for trait, part_damage in PILOT_FIGHTERDAMAGE_UNSCALED_MODIFIER_DICT.items()
}

# </editor-fold>

# <editor-fold desc="Extraordinary Species Rules">
# tag used if species self-destructs on conquest
TAG_DESTROYED_ON_CONQUEST = "DESTROYED_ON_CONQUEST"

# some species have a fixed population
SPECIES_FIXED_POPULATION = {
    # species_name: fixed_population_size
    SP_LEMBALALAM: 5.0,
}

# techs that are unlocked if conquering a planet of a species
SPECIES_TECH_UNLOCKS = {
    # species: [tech1, tech2, ...]
    SP_LEMBALALAM: [GRO_LIFE_CYCLE, SPY_STEALTH_1, SPY_STEALTH_2]
}
# </editor-fold>

# </editor-fold>


# <editor-fold desc="Buildings">
# <editor-fold desc="Supply">
# Please see the Note at top of this file regarding PlanetSize-Dependent-Lookup
# building supply bonuses are keyed by planet size; key -1 stands for any planet size
building_supply = {
    # Palace also gives a bonus, but it is destroyed when being conquered, so we wouldn't get it.
    "BLD_MEGALITH": {
        -1: 2,
    },
    "BLD_SPACE_ELEVATOR": {
        fo.planetSize.tiny: 1,
        fo.planetSize.small: 2,
        fo.planetSize.medium: 3,
        fo.planetSize.large: 4,
        fo.planetSize.huge: 5,
        fo.planetSize.gasGiant: 3,
    },
}
# </editor-fold>

# </editor-fold>


# <editor-fold desc="Shipdesign/-parts">
PART_KRILL_SPAWNER = "SP_KRILL_SPAWNER"


HULL_EXCLUDED_SHIP_PART_CLASSES = {"SH_COLONY_BASE": (fo.shipPartClass.fuel, fo.shipPartClass.speed)}


# <editor-fold desc="Allowed Combat Targets">
# TODO: Inherit from enum.Flag after switch to Python 3.6+
class CombatTarget:
    NONE = 0
    SHIP = 1 << 0
    PLANET = 1 << 1
    FIGHTER = 1 << 2
    ANY = SHIP | PLANET | FIGHTER

    # map from weapon to allowed targets
    PART_ALLOWED_TARGETS = {
        "SR_WEAPON_0_1": FIGHTER,
        "SR_WEAPON_1_1": SHIP | PLANET,
        "SR_WEAPON_2_1": SHIP | PLANET,
        "SR_WEAPON_3_1": SHIP | PLANET,
        "SR_WEAPON_4_1": SHIP | PLANET,
        "SR_ARC_DISRUPTOR": ANY,
        SR_FLUX_LANCE: SHIP | PLANET,
        "SR_SPINAL_ANTIMATTER": SHIP | PLANET,
        "FT_HANGAR_0": NONE,
        "FT_HANGAR_1": FIGHTER,
        "FT_HANGAR_2": SHIP | FIGHTER,
        "FT_HANGAR_3": SHIP,
        "FT_HANGAR_4": SHIP | PLANET,
        # monster weapons
        "SR_GRAV_PULSE": ANY,
        "SR_ICE_BEAM": ANY,
        "SR_JAWS": SHIP | PLANET,
        "SR_PLASMA_DISCHARGE": ANY,
        "SR_SPINES": SHIP | PLANET,
        "SR_TENTACLE": ANY,  # note that SHIPs are primary targets
        "SR_THRASHING_BODY": SHIP | FIGHTER,
        "FT_HANGAR_KRILL": SHIP | FIGHTER,
    }


# </editor-fold>


# <editor-fold desc="Effect Scripting for Shipdesigns">
# <editor-fold desc="Tokens">
# known tokens the AI can handle
REPAIR_PER_TURN = "REPAIR_PER_TURN"
FUEL_PER_TURN = "FUEL_PER_TURN"
STEALTH_MODIFIER = "STEALTH_MODIFIER"
ASTEROID_STEALTH = "ASTEROID_STEALTH"
SOLAR_STEALTH = "SOLAR_STEALTH"
SPEED = "SPEED"
FUEL = "FUEL"
FUEL_EFFICIENCY = "FUEL_EFFICIENCY"
SHIELDS = "SHIELDS"
STRUCTURE = "STRUCTURE"
DETECTION = "DETECTION"  # do only specify for hulls if irregular detection
ORGANIC_GROWTH = "ORGANIC_GROWTH"  # structure for value is (per_turn, maximum)
STACKING_RULES = "STACKING_RULES"  # expects a list of stacking rules
# <editor-fold desc="Stacking rules">
NO_EFFECT_WITH_CLOAKS = "NO_EFFECT_WITH_CLOAKS"  # part does not stack with cloak parts
# </editor-fold>
# </editor-fold>

BASE_DETECTION = 25

TECH_EFFECTS = {
    # "TECHNAME": { Token1: Value1, Token2: Value2, ...}
    "SHP_REINFORCED_HULL": {STRUCTURE: 5 * SHIP_STRUCTURE_FACTOR},
    "SHP_BASIC_DAM_CONT": {REPAIR_PER_TURN: SHIP_STRUCTURE_FACTOR},
    "SHP_FLEET_REPAIR": {REPAIR_PER_TURN: (STRUCTURE, 0.1)},  # 10% of max structure
    "SHP_ADV_DAM_CONT": {REPAIR_PER_TURN: (STRUCTURE, 0.1)},  # 10% of max structure
    "SHP_INTSTEL_LOG": {SPEED: 20},  # technically not correct, but as approximation good enough...
    "GRO_ENERGY_META": {FUEL: 1},
}

DEFAULT_FUEL_EFFICIENCY = 1
HULL_TAG_EFFECTS = {
    "GREAT_FUEL_EFFICIENCY": {
        FUEL_EFFICIENCY: 4,
    },
    "GOOD_FUEL_EFFICIENCY": {
        FUEL_EFFICIENCY: 2,
    },
    "AVERAGE_FUEL_EFFICIENCY": {
        FUEL_EFFICIENCY: 1,
    },
    "BAD_FUEL_EFFICIENCY": {
        FUEL_EFFICIENCY: 0.6,
    },
}

HULL_EFFECTS = {
    # "HULLNAME": { Token1: Value1, Token2: Value2, ...}
    # Spatial Flux line
    "SH_SPACE_FLUX_BUBBLE": {
        STEALTH_MODIFIER: -30,
    },
    "SH_SPACE_FLUX_COMPOSITE": {
        STEALTH_MODIFIER: -30,
    },
    "SH_SPATIAL_FLUX": {
        STEALTH_MODIFIER: -30,
    },
    # Robotic line
    "SH_ROBOTIC": {
        REPAIR_PER_TURN: 2 * SHIP_STRUCTURE_FACTOR,
    },
    "SH_NANOROBOTIC": {
        REPAIR_PER_TURN: (STRUCTURE, 1),
    },  # 100% of max structure
    "SH_LOGISTICS_FACILITATOR": {
        REPAIR_PER_TURN: (STRUCTURE, 1),
    },  # 100% of max structure
    # Asteroid line
    "SH_SMALL_ASTEROID": {
        ASTEROID_STEALTH: 20,
    },
    "SH_ASTEROID": {
        ASTEROID_STEALTH: 20,
    },
    "SH_HEAVY_ASTEROID": {
        ASTEROID_STEALTH: 20,
    },
    "SH_SMALL_CAMOUFLAGE_ASTEROID": {
        ASTEROID_STEALTH: 20,
    },
    "SH_CAMOUFLAGE_ASTEROID": {
        ASTEROID_STEALTH: 40,
    },
    "SH_CRYSTALLIZED_ASTEROID": {
        ASTEROID_STEALTH: 20,
    },
    "SH_MINIASTEROID_SWARM": {
        ASTEROID_STEALTH: 20,
        SHIELDS: 5 * SHIP_WEAPON_DAMAGE_FACTOR,
    },
    "SH_SCATTERED_ASTEROID": {
        ASTEROID_STEALTH: 40,
        SHIELDS: 3 * SHIP_WEAPON_DAMAGE_FACTOR,
    },
    # Organic line
    "SH_ORGANIC": {
        REPAIR_PER_TURN: 2 * SHIP_STRUCTURE_FACTOR,
        FUEL_PER_TURN: 0.2,
        ORGANIC_GROWTH: (0.2 * SHIP_STRUCTURE_FACTOR, 3 * SHIP_STRUCTURE_FACTOR),
    },
    "SH_ENDOMORPHIC": {
        DETECTION: 50,
        ORGANIC_GROWTH: (0.5 * SHIP_STRUCTURE_FACTOR, 15 * SHIP_STRUCTURE_FACTOR),
    },
    "SH_SYMBIOTIC": {
        REPAIR_PER_TURN: 2 * SHIP_STRUCTURE_FACTOR,
        FUEL_PER_TURN: 0.2,
        DETECTION: 40,
        ORGANIC_GROWTH: (0.2 * SHIP_STRUCTURE_FACTOR, 10 * SHIP_STRUCTURE_FACTOR),
    },
    "SH_PROTOPLASMIC": {
        REPAIR_PER_TURN: 2 * SHIP_STRUCTURE_FACTOR,
        FUEL_PER_TURN: 0.2,
        DETECTION: 40,
        ORGANIC_GROWTH: (0.5 * SHIP_STRUCTURE_FACTOR, 25 * SHIP_STRUCTURE_FACTOR),
    },
    "SH_ENDOSYMBIOTIC": {
        REPAIR_PER_TURN: 2 * SHIP_STRUCTURE_FACTOR,
        FUEL_PER_TURN: 0.2,
        DETECTION: 40,
        ORGANIC_GROWTH: (0.5 * SHIP_STRUCTURE_FACTOR, 15 * SHIP_STRUCTURE_FACTOR),
    },
    "SH_RAVENOUS": {
        DETECTION: 75,
        ORGANIC_GROWTH: (0.5 * SHIP_STRUCTURE_FACTOR, 20 * SHIP_STRUCTURE_FACTOR),
    },
    "SH_BIOADAPTIVE": {
        REPAIR_PER_TURN: (STRUCTURE, 1),
        FUEL_PER_TURN: 0.2,
        DETECTION: 75,
        ORGANIC_GROWTH: (0.5 * SHIP_STRUCTURE_FACTOR, 25 * SHIP_STRUCTURE_FACTOR),
    },
    "SH_SENTIENT": {
        REPAIR_PER_TURN: 2 * SHIP_STRUCTURE_FACTOR,
        FUEL_PER_TURN: 0.2,
        DETECTION: 90,
        ORGANIC_GROWTH: (1 * SHIP_STRUCTURE_FACTOR, 45 * SHIP_STRUCTURE_FACTOR),
        STEALTH_MODIFIER: 20,
    },
    # Energy Line
    "SH_SOLAR": {
        SOLAR_STEALTH: 120,
        FUEL_PER_TURN: (FUEL, 1),
    },  # 100% of fuel
}

PART_EFFECTS = {
    # "PARTNAME": { Token1: Value1, Token2: Value2, ...}
    "SH_MULTISPEC": {
        SOLAR_STEALTH: 60,
    },
    "FU_TRANSPATIAL_DRIVE": {},  # not supported yet
    "FU_RAMSCOOP": {
        FUEL_PER_TURN: 0.1,
    },
    "FU_ZERO_FUEL": {
        FUEL_PER_TURN: (FUEL, 1),
    },  # 100% of fuel
    "SP_DISTORTION_MODULATOR": {},  # not supported yet
    "SH_ROBOTIC_INTERFACE_SHIELDS": {},  # not supported yet
    PART_KRILL_SPAWNER: {
        STEALTH_MODIFIER: 40,
        STACKING_RULES: [NO_EFFECT_WITH_CLOAKS],
    },
}

# </editor-fold>

# </editor-fold>



from focs._effects import (
    CurrentTurn,
    EffectsGroup,
    LocalCandidate,
    MinOf,
    NamedReal,
    OwnedBy,
    Planet,
    Population,
    SetMaxTroops,
    SetTroops,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER, TROOPS_PER_POP
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    TARGET_AFTER_SCALING_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
)

Tech(
    name="DEF_GARRISON_1",
    description="DEF_GARRISON_1_DESC",
    short_description="TROOPS_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=9 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_ROOT_DEFENSE"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            stackinggroup="GARRISON_1_TROOPS_STACK",
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetMaxTroops(value=Value + NamedReal(name="GARRISON_1_MAX_TROOPS_FLAT", value=6)),
        ),
    ],
    graphic="icons/tech/troops.png",
)

Tech(
    name="DEF_GARRISON_2",
    description="DEF_GARRISON_2_DESC",
    short_description="TROOPS_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=25 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_GARRISON_1"],
    unlock=[
        Item(type=UnlockPolicy, name="PLC_CHECKPOINTS"),
        Item(type=UnlockPolicy, name="PLC_INSURGENCY"),
    ],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & (LocalCandidate.LastTurnConquered < CurrentTurn)
            & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetTroops(
                value=MinOf(
                    float, Value(Target.MaxTroops), Value + NamedReal(name="GARRISON_2_TROOPREGEN_FLAT", value=1)
                )
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Population(low=0.01),
            stackinggroup="GARRISON_2_TROOPS_STACK",
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
            effects=SetMaxTroops(
                value=Value
                + Target.Population * NamedReal(name="GARRISON_2_MAXTROOPS_PERPOP", value=2 * TROOPS_PER_POP)
            ),
        ),
    ],
    graphic="icons/tech/troops.png",
)

Tech(
    name="DEF_GARRISON_3",
    description="DEF_GARRISON_3_DESC",
    short_description="TROOPS_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=84 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_GARRISON_2"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & (LocalCandidate.LastTurnConquered < CurrentTurn)
            & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetTroops(
                value=MinOf(
                    float, Value(Target.MaxTroops), Value + NamedReal(name="GARRISON_3_TROOPREGEN_FLAT", value=2)
                )
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner),
            stackinggroup="GARRISON_3_TROOPS_STACK",
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetMaxTroops(value=Value + NamedReal(name="GARRISON_3_MAX_TROOPS_FLAT", value=18)),
        ),
    ],
    graphic="icons/tech/troops.png",
)

Tech(
    name="DEF_GARRISON_4",
    description="DEF_GARRISON_4_DESC",
    short_description="TROOPS_SHORT_DESC",
    category="DEFENSE_CATEGORY",
    researchcost=216 * TECH_COST_MULTIPLIER,
    researchturns=9,
    tags=["PEDIA_DEFENSE_CATEGORY"],
    prerequisites=["DEF_GARRISON_3"],
    unlock=Item(type=UnlockPolicy, name="PLC_MARTIAL_LAW"),
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & (LocalCandidate.LastTurnConquered < CurrentTurn)
            & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
            effects=SetTroops(
                value=MinOf(
                    float, Value(Target.MaxTroops), Value + NamedReal(name="GARRISON_4_TROOPREGEN_FLAT", value=4)
                )
            ),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Population(low=0.01),
            stackinggroup="GARRISON_4_TROOPS_STACK",
            accountinglabel="DEF_TECH_ACCOUNTING_LABEL",
            priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
            effects=SetMaxTroops(
                value=Value
                + Target.Population * NamedReal(name="GARRISON_4_MAX_TROOPS_PERPOP", value=3 * TROOPS_PER_POP)
            ),
        ),
    ],
    graphic="icons/tech/troops.png",
)



from colonization.nest_rating import get_nest_rating, special_is_nest
from colonization.planet_supply import get_planet_supply, update_planet_supply
from colonization.rate_pilots import rate_piloting, rate_planetary_piloting



from buildings import Shipyard
from common.fo_typing import BuildingName, PlanetId
from empire.pilot_rating import best_pilot_rating
from empire.survey_lock import survey_universe_lock
from freeorion_tools.caching import cache_for_current_turn


@survey_universe_lock
def get_best_pilot_facilities(facility: BuildingName) -> frozenset[PlanetId]:
    """Gives list of planets with the best available pilots have the given ship building facility"""
    # TBD: this is only for battle ships and ignores things like fuel and vision range.
    # For scouts, vision and fuel would be more important
    best_pilot_facilities = _get_facilities().get("WEAPONS_%.1f" % best_pilot_rating(), {})

    return best_pilot_facilities.get(facility, set())


@cache_for_current_turn
def _get_facilities() -> dict[str, dict[BuildingName, set[PlanetId]]]:
    return {}


def set_building_locations(
    weapons_grade: str,
    buildings_here: set[BuildingName],
    pid: PlanetId,
):
    this_grade_facilities = _get_facilities().setdefault(weapons_grade, {})
    for building in buildings_here:
        try:
            Shipyard(building)
        except ValueError:
            continue
        this_grade_facilities.setdefault(building, set()).add(pid)



import pytest
from pytest import raises

from freeorion_tools import ReadOnlyDict

dict_content = {
    1: -1,
    1.09: 1.1,
    "abc": "xyz",
    (1, 2, 3): ("a", 1, (1, 2)),
}


def test_readonly_dict_created_from_dict_behaves_the_same():
    test_dict = ReadOnlyDict(dict_content)
    assert set(test_dict.keys()) == set(dict_content.keys())
    assert set(test_dict.values()) == set(dict_content.values())
    assert set(test_dict.items()) == set(dict_content.items())
    assert len(test_dict) == len(dict_content)


def test_readonly_dict_created_from_dict_iterates_over_same_keys():
    test_dict = ReadOnlyDict(dict_content)
    assert set(test_dict) == set(dict_content)


def test_readonly_dict_get_existing_key_works_same_as_for_dict():
    test_dict = ReadOnlyDict(dict_content)
    # check for membership checks and retrieval
    for key, value in dict_content.items():
        assert key in test_dict
        assert test_dict[key] == value
        assert test_dict.get(key, -99999) == value


def test_readonly_dict_get_non_existiong_key_works_same_as_for_dict():
    test_dict = ReadOnlyDict(dict_content)
    # check correct functionality if keys not in dict
    assert "INVALID_KEY" not in test_dict
    assert test_dict.get("INVALID_KEY", -99999) == -99999
    with pytest.raises(KeyError, match="'INVALID_KEY'"):
        # noinspection PyStatementEffect
        test_dict["INVALID_KEY"]
        pytest.fail("Invalid key lookup didn't raise a KeyError")


def test_readonly_dict_evaluated_to_bool_same_as_collection():
    test_dict = ReadOnlyDict(dict_content)
    assert bool(test_dict)
    assert test_dict
    empty_dict = ReadOnlyDict()
    assert not len(empty_dict)
    assert not empty_dict


def test_readonly_dict_created_from_dict_can_be_converted_to_the_same_dict():
    read_only_dict = ReadOnlyDict(dict_content)
    normal_dict = dict(read_only_dict)
    assert len(normal_dict) == len(dict_content)
    assert set(normal_dict.items()) == set(dict_content.items())


def test_readonly_dict_remove_value_raises_error():
    test_dict = ReadOnlyDict(dict_content)
    with raises(TypeError):
        del test_dict[1]
        raise AssertionError("Can delete items from the dict.")


def test_readonly_dict_set_value_raises_error():
    test_dict = ReadOnlyDict(dict_content)
    with raises(TypeError, match="'ReadOnlyDict' object does not support item assignment"):
        test_dict["INVALID_KEY"] = 1
        pytest.fail("Can add items to the dict.")
    assert "INVALID_KEY" not in test_dict


def test_readonly_create_from_dict_with_hashable_values_raises_error():
    with raises(TypeError, match="unhashable type: 'list'"):
        ReadOnlyDict({1: [1]})
        pytest.fail("Can store mutable items in dict")



from focs._effects import Capital, EffectsGroup, GiveEmpireTech, IsSource, Planet, Target, Turn, Unowned
from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import NARROW_EP, TUNDRA_NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.fuel import GREAT_FUEL
from species.species_macros.happiness import BAD_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import AVERAGE_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import BAD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import GOOD_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.telepathic import PRECOGNITIVE_DETECTION
from species.species_macros.troops import BAD_DEFENSE_TROOPS
from species.species_macros.weapons import GOOD_WEAPONS

Species(
    name="SP_FULVER",
    description="SP_FULVER_DESC",
    gameplay_description="SP_FULVER_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "LITHIC",
        "GOOD_STOCKPILE",
        "GOOD_WEAPONS",
        "BAD_RESEARCH",
        "BAD_HAPPINESS",
        "AVERAGE_SUPPLY",
        "GREAT_FUEL",
        "TELEPATHIC",
        "PEDIA_LITHIC_SPECIES_CLASS",
        "PEDIA_TELEPATHIC_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_STOCKPILE",
    likes=[
        "BLD_HYPER_DAM",
        "BLD_STARGATE",
        "CRYSTALS_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "FOCUS_INDUSTRY",
        "FOCUS_RESEARCH",
        "FOCUS_STOCKPILE",
        "GAIA_SPECIAL",
        "MIMETIC_ALLOY_SPECIAL",
        "PLC_ALGORITHMIC_RESEARCH",
        "PLC_COLONIZATION",
        "PLC_CONFEDERATION",
        "PLC_DREAM_RECURSION",
        "PLC_ENGINEERING",
        "PLC_EXPLORATION",
        "PLC_INSURGENCY",
        "PLC_LIBERTY",
        "PLC_MODERATION",
        "PLC_NECESSITY",
        "PLC_RACIAL_PURITY",
        "PLC_STOCKPILE_LIQUIDATION",
        "RESONANT_MOON_SPECIAL",
        "SUCCULENT_BARNACLES_SPECIAL",
    ],
    dislikes=[
        "FOCUS_GROWTH",
        "FOCUS_INFLUENCE",
        "FOCUS_LOGISTICS",
        "FOCUS_PROTECTION",
        "ANCIENT_RUINS_DEPLETED_SPECIAL",
        "BLD_GAS_GIANT_GEN",
        "BLD_INDUSTRY_CENTER",
        "BLD_REGIONAL_ADMIN",
        "BLD_SHIPYARD_ENRG_COMP",
        "BLD_SHIPYARD_ENRG_SOLAR",
        "BLD_SOL_ORB_GEN",
        "FORTRESS_SPECIAL",
        "KRAKEN_NEST_SPECIAL",
        "PLC_BUREAUCRACY",
        "PLC_CHECKPOINTS",
        "PLC_COLONIALISM",
        "PLC_ENVIRONMENTALISM",
        "PLC_FEUDALISM",
        "PLC_INTERSTELLAR_INFRA",
        "PLC_METROPOLES",
        "PLC_SYSTEM_INFRA",
        "PLC_TECHNOCRACY",
        "PLC_VASSALIZATION",
        "SUPERCONDUCTOR_SPECIAL",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *BAD_RESEARCH,
        *AVERAGE_INFLUENCE,
        *GOOD_STOCKPILE,
        *AVERAGE_POPULATION,
        *BAD_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_FULVER"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *GOOD_WEAPONS,
        *GREAT_FUEL,
        *PRECOGNITIVE_DETECTION(2),
        EffectsGroup(
            description="FULVER_STARTING_UNLOCKS_DESC",
            scope=IsSource,
            activation=Turn(high=1),
            effects=[GiveEmpireTech(name="PRO_GENERIC_SUPPLIES", empire=Target.Owner)],
        ),
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        NARROW_EP,
        *STANDARD_SHIP_SHIELDS,
        EffectsGroup(
            scope=IsSource, activation=Planet() & ~Unowned & Capital, effects=GiveEmpirePolicy(name="PLC_CONFEDERATION")
        ),
    ],
    environments=TUNDRA_NARROW_EP,
    graphic="icons/species/insectoid-01.png",
)



import freeOrionAIInterface as fo

import AIDependencies
import PolicyAI
from buildings import BuildingType, iterate_buildings_types
from colonization.colony_score import debug_rating
from EnumsAI import FocusType
from freeorion_tools import (
    get_game_rule_int,
    get_named_int,
    get_named_real,
    get_species_stability,
)
from freeorion_tools.caching import cache_for_current_turn
from PlanetUtilsAI import dislike_factor
from turn_state import get_empire_planets_by_species, have_worldtree, luxury_resources
from universe.system_network import within_n_jumps

_size_modifier = {
    fo.planetSize.tiny: get_game_rule_int("RULE_TINY_SIZE_STABILITY", 2),
    fo.planetSize.small: get_game_rule_int("RULE_SMALL_SIZE_STABILITY", 1),
    fo.planetSize.medium: get_game_rule_int("RULE_MEDIUM_SIZE_STABILITY", 0),
    fo.planetSize.large: get_game_rule_int("RULE_LARGE_SIZE_STABILITY", -1),
    fo.planetSize.huge: get_game_rule_int("RULE_HUGE_SIZE_STABILITY", -2),
    fo.planetSize.gasGiant: get_game_rule_int("RULE_GAS_GIANT_SIZE_STABILITY", 0),
    fo.planetSize.asteroids: 0,  # no rule for asteroids yet(?)
}

_environment_modifier = {
    fo.planetEnvironment.good: get_game_rule_int("RULE_GOOD_ENVIRONMENT_STABILITY", 2),
    fo.planetEnvironment.adequate: get_game_rule_int("RULE_ADEQUATE_ENVIRONMENT_STABILITY", 1),
    fo.planetEnvironment.poor: get_game_rule_int("RULE_POOR_ENVIRONMENT_STABILITY", 0),
    fo.planetEnvironment.hostile: get_game_rule_int("RULE_HOSTILE_ENVIRONMENT_STABILITY", -1),
    # That is used by the script. It doesn't matter in practice, AI shouldn't try to settle uninhabitable planets.
    fo.planetEnvironment.uninhabitable: get_game_rule_int("RULE_HOSTILE_ENVIRONMENT_STABILITY", -1),
}


def calculate_stability(planet: fo.planet, species: fo.species) -> float:
    """
    Calculate the focus-independent stability species should have on planet.
    Distance to capital never give negatives, since we could build a regional admin, if it would.
    Supply-connection to nearest regional admin still TBD.
    """
    baseline = fo.getGameRules().getDouble("RULE_BASELINE_PLANET_STABILITY")
    species_mod = get_species_stability(species.name)
    home_bonus = AIDependencies.STABILITY_HOMEWORLD_BONUS if planet.id in species.homeworlds else 0.0
    policies = _evaluate_policies(species)
    specials = _evaluate_specials(planet, species)
    buildings = _evaluate_buildings(planet, species)
    xenophobia = _evaluate_xenophobia(planet, species)
    administration = _evaluate_administration(planet, species)
    rules = _size_modifier[planet.size] + _environment_modifier[planet.environmentForSpecies(species.name)]
    result = baseline + species_mod + home_bonus + policies + specials + buildings + xenophobia + administration + rules
    # missing: supply connection check, artisan bonus, anything else?
    debug_rating(
        f"stability of {species.name} on {planet} would be {result:.1f} (base={baseline:.1f}, "
        f"species={species_mod:.1f}, specials={specials:.1f}, home={home_bonus:.1f}, policies={policies:.1f}, "
        f"buildings={buildings:.1f}, xeno={xenophobia:.1f}, admin={administration:.1f}), rules={rules}"
    )
    return result


@cache_for_current_turn
def _evaluate_policies(species: fo.species) -> float:
    empire = fo.getEmpire()
    like_value = AIDependencies.STABILITY_PER_LIKED_FOCUS
    dislike_value = like_value * dislike_factor()
    result = sum(like_value for p in empire.adoptedPolicies if p in species.likes)
    result -= sum(dislike_value for p in empire.adoptedPolicies if p in species.dislikes)
    if PolicyAI.bureaucracy in empire.adoptedPolicies:
        result += get_named_real("PLC_BUREAUCRACY_STABILITY_FLAT")
    if PolicyAI.diversity in empire.adoptedPolicies:
        current_species = get_empire_planets_by_species()
        # The evaluated planet may add another species
        num_species = len(current_species) + (1 if species not in current_species else 0)
        diversity_value = num_species - get_named_int("PLC_DIVERSITY_THRESHOLD")
        diversity_scaling = get_named_real("PLC_DIVERSITY_SCALING")
        result += diversity_value * diversity_scaling
    if PolicyAI.capital_markets in empire.adoptedPolicies:
        for special, planets in luxury_resources().items():
            if special in species.likes and any(planet.focus == FocusType.FOCUS_INFLUENCE for planet in planets):
                result += get_named_real("CAPITAL_MARKETS_INFLUENCE_BONUS_SCALING")
    # TBD: add conformance, indoctrination, etc. when the AI learns to use them
    return result


def _evaluate_specials(planet: fo.planet, species: fo.species) -> float:
    universe = fo.getUniverse()
    system = universe.getSystem(planet.systemID)
    result = 0.0
    for pid in system.planetIDs:
        if pid == planet.id:
            value = AIDependencies.STABILITY_PER_LIKED_SPECIAL_ON_PLANET
            eval_planet = planet
        else:
            value = AIDependencies.STABILITY_PER_LIKED_SPECIAL_IN_SYSTEM
            eval_planet = universe.getPlanet(pid)
        for special in eval_planet.specials:
            if special in species.likes:
                result += value
            if special in species.dislikes:
                result -= value
    if have_worldtree() and not AIDependencies.not_affect_by_special(AIDependencies.WORLDTREE_SPECIAL, species.name):
        result += AIDependencies.STABILITY_BY_WORLDTREE
    gaia = AIDependencies.GAIA_SPECIAL
    if gaia in planet.specials and not AIDependencies.not_affect_by_special(gaia, species):
        result += 5  # TBD add named real
    return result


@cache_for_current_turn
def _count_building(planet: fo.planet) -> dict[str, tuple[int, int, int]]:
    """Returns Mapping from BuildingType to number of buildings on planet, in system and elsewhere."""
    universe = fo.getUniverse()
    system = universe.getSystem(planet.systemID)
    planet_pid = {planet.id}
    system_pids = {pid for pid in system.planetIDs if pid != planet.id}
    # TODO: add all buildings to BuildingType, so we get them all here
    result = {}
    for building_type in iterate_buildings_types():
        # So far we only consider conquering / settling this planet. By the time we have it,
        # queued buildings are likely finished as well, so we consider them here already.
        all_pids = building_type.built_or_queued_at()
        result[building_type.value] = (
            len(all_pids & planet_pid),
            len(all_pids & system_pids),
            len(all_pids - system_pids - planet_pid),
        )
    return result


def _evaluate_buildings(planet: fo.planet, species: fo.species) -> float:
    result = 0.0
    for name, numbers in _count_building(planet).items():
        if name in species.likes:
            result += (
                numbers[0] * AIDependencies.STABILITY_PER_LIKED_BUILDING_ON_PLANET
                + numbers[1] * AIDependencies.STABILITY_PER_LIKED_BUILDING_IN_SYSTEM
                + numbers[2] ** 0.5 * AIDependencies.STABILITY_BASE_LIKED_BUILDING_ELSEWHERE
            )
        elif name in species.dislikes:
            result -= dislike_factor() * (
                numbers[0] * AIDependencies.STABILITY_PER_LIKED_BUILDING_ON_PLANET
                + numbers[1] * AIDependencies.STABILITY_PER_LIKED_BUILDING_IN_SYSTEM
                + numbers[2] ** 0.5 * AIDependencies.STABILITY_BASE_LIKED_BUILDING_ELSEWHERE
            )
    return result


def _evaluate_xenophobia(planet, species) -> float:
    if AIDependencies.Tags.XENOPHOBIC not in species.tags:
        return 0.0
    universe = fo.getUniverse()
    max_jumps = get_named_int("XENOPHOBIC_MAX_JUMPS")
    relevant_systems = within_n_jumps(planet.systemID, max_jumps)
    penalty_perjump = get_named_real("XENOPHOBIC_TARGET_STABILITY_PERJUMP")
    result = 0.0
    for sys_id in relevant_systems:
        system = universe.getSystem(sys_id)
        for pid in system.planetIDs:
            planet_species = universe.getPlanet(pid).speciesName
            # TODO Do not count owned, different species planets when Racial Purity is adopted. AI doesn't adopt Racial Purity yet.
            if planet_species not in ("SP_EXOBOT", species.name, ""):
                result += penalty_perjump * (1 + max_jumps - universe.jumpDistance(planet.systemID, pid))
    return result


def _evaluate_administration(planet: fo.planet, species: fo.species) -> float:
    result = 0.0
    palace = BuildingType.PALACE
    universe = fo.getUniverse()
    if AIDependencies.Tags.INDEPENDENT not in species.tags:
        admin_systems = BuildingType.REGIONAL_ADMIN.built_or_queued_at_sys() | palace.built_or_queued_at_sys()
        jumps_to_admin = min((universe.jumpDistance(planet.systemID, admin) for admin in admin_systems), default=99)
        # cap at 5, if it is 6 or more, we could build a Regional Admin on this planet
        # TODO: check if the planet would be supply-connected to the nearest administration
        # disconnected gives namedReal("DISCONNECTED_FROM_CAPITAL_AND_REGIONAL_ADMIN_STABILITY_PENALTY")
        result += 5 - min(jumps_to_admin, 5)
    if palace.built_at():  # bonus is only given the capital actually contains the palace
        # When rebuilding a palace, the planet only becomes the capital the turn after the palace is finished.
        # So we cannot use fo.getEmpire().capitalID here, but for the calculation we do consider it the capital.
        capital = universe.getPlanet(list(palace.built_at())[0])
        if species.name == capital.speciesName:
            result += 5.0
    return result



from inspect import getdoc
from typing import Any, Optional

from stub_generator.interface_inspector.inspection_helpers import (
    _get_member_info,
    _getmembers,
)


class ClassInfo:
    def __init__(self, name: str, attributes: dict[str, Any], doc: Optional[str], parents: list[str]):
        self.name = name
        self.attributes = attributes
        self.doc = doc
        self.parents = parents

    def __repr__(self):
        return f"ClassInfo({self.name!r}, {self.attributes!r}, {self.doc!r}, {self.parents!r})"


def inspect_class(class_name, obj):
    parents = obj.mro()[1:-2]
    parent_attrs = sum((dir(parent) for parent in obj.mro()[1:]), [])

    attrs = {}

    for attr_name, member in _getmembers(obj):
        if attr_name not in parent_attrs + ["__module__", "__instance_size__"]:
            attrs[attr_name] = _get_member_info(f"{class_name}.{attr_name}", member)
    return ClassInfo(class_name, attrs, getdoc(obj), [str(parent.__name__) for parent in parents])



from focs._tech import *

Tech(
    name="SHP_ROOT_ARMOR",
    description="SHP_ROOT_ARMOR_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="SHIP_PARTS_CATEGORY",
    researchcost=1,
    researchturns=1,
    tags=["PEDIA_ARMOR_PART_TECHS"],
    unlock=Item(type=UnlockShipPart, name="AR_STD_PLATE"),
    graphic="icons/tech/armor_plating.png",
)
